# Copyright (C) 2005-2012 Aaron Bentley <abentley@panoramicfeedback.com>
#                         Chris Ball <cjb@laptop.org>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

import base

UserError = base.UserError
UsageError = base.UsageError
UnknownCommand = base.UnknownCommand
get_command = base.get_command
get_command_class = base.get_command_class
commands = base.commands
Option = base.Option
Argument = base.Argument
Command = base.Command
InputOutput = base.InputOutput
StdInputOutput = base.StdInputOutput
StringInputOutput = base.StringInputOutput
UnconnectedStorageGetter = base.UnconnectedStorageGetter
StorageCallbacks = base.StorageCallbacks
UserInterface = base.UserInterface

__all__ = [UserError, UsageError, UnknownCommand,
           get_command, get_command_class, commands,
           Option, Argument, Command,
           InputOutput, StdInputOutput, StringInputOutput,
           StorageCallbacks, UnconnectedStorageGetter,
           UserInterface]



# Copyright (C) 2005-2012 Aaron Bentley <abentley@panoramicfeedback.com>
#                         Chris Ball <cjb@laptop.org>
#                         Gianluca Montecchi <gian@grys.it>
#                         Marien Zwart <marien.zwart@gmail.com>
#                         Thomas Gerigk <tgerigk@gmx.de>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

import libbe
import libbe.command
import libbe.command.util


class Remove (libbe.command.Command):
    """Remove (delete) a bug and its comments

    >>> import sys
    >>> import libbe.bugdir
    >>> bd = libbe.bugdir.SimpleBugDir(memory=False)
    >>> io = libbe.command.StringInputOutput()
    >>> io.stdout = sys.stdout
    >>> ui = libbe.command.UserInterface(io=io)
    >>> ui.storage_callbacks.set_storage(bd.storage)
    >>> cmd = Remove(ui=ui)

    >>> print bd.bug_from_uuid('b').status
    closed
    >>> ret = ui.run(cmd, args=['/b'])
    Removed bug abc/b
    >>> bd.flush_reload()
    >>> try:
    ...     bd.bug_from_uuid('b')
    ... except libbe.bugdir.NoBugMatches:
    ...     print 'Bug not found'
    Bug not found
    >>> ui.cleanup()
    >>> bd.cleanup()
    """
    name = 'remove'

    def __init__(self, *args, **kwargs):
        libbe.command.Command.__init__(self, *args, **kwargs)
        self.args.extend([
                libbe.command.Argument(
                    name='bug-id', metavar='BUG-ID', default=None,
                    repeatable=True,
                    completion_callback=libbe.command.util.complete_bug_id),
                ])

    def _run(self, **params):
        bugdirs = self._get_bugdirs()
        user_ids = []
        for bug_id in params['bug-id']:
            bugdir,bug,comment = (
                libbe.command.util.bugdir_bug_comment_from_user_id(
                    bugdirs, bug_id))
            user_ids.append(bug.id.user())
            bugdir.remove_bug(bug)
        if len(user_ids) == 1:
            print >> self.stdout, 'Removed bug %s' % user_ids[0]
        else:
            print >> self.stdout, 'Removed bugs %s' % ', '.join(user_ids)
        return 0

    def _long_help(self):
        return """
Remove (delete) existing bugs.  Use with caution: if you're not using
a revision control system, there may be no way to recover the lost
information.  You should use this command, for example, to get rid of
blank or otherwise mangled bugs.
"""



# Copyright (C) 2009-2012 Chris Ball <cjb@laptop.org>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

"""Define the Version Controlled System (VCS)-based
:py:class:`~libbe.storage.base.Storage` and
:py:class:`~libbe.storage.base.VersionedStorage` implementations.

There is a base class (:py:class:`~libbe.storage.vcs.VCS`) translating 
Storage language to VCS language, and a number of `VCS` implementations:

* :py:class:`~libbe.storage.vcs.arch.Arch`
* :py:class:`~libbe.storage.vcs.bzr.Bzr`
* :py:class:`~libbe.storage.vcs.darcs.Darcs`
* :py:class:`~libbe.storage.vcs.git.Git`
* :py:class:`~libbe.storage.vcs.hg.Hg`

The base `VCS` class also serves as a filesystem Storage backend (not
versioning) in the event that a user has no VCS installed.
"""

import base

set_preferred_vcs = base.set_preferred_vcs
vcs_by_name = base.vcs_by_name
detect_vcs = base.detect_vcs
installed_vcs = base.installed_vcs

__all__ = [set_preferred_vcs, vcs_by_name, detect_vcs, installed_vcs]



# Copyright (C) 2005-2012 Aaron Bentley <abentley@panoramicfeedback.com>
#                         Ben Finney <benf@cybersource.com.au>
#                         Chris Ball <cjb@laptop.org>
#                         Gianluca Montecchi <gian@grys.it>
#                         Marien Zwart <marien.zwart@gmail.com>
#                         Michel Alexandre Salim <salimma@fedoraproject.org>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

"""Bazaar_ (bzr) backend.

.. _Bazaar: http://bazaar.canonical.com/
"""

try:
    import bzrlib
    import bzrlib.branch
    import bzrlib.builtins
    import bzrlib.config
    import bzrlib.errors
    import bzrlib.option
except ImportError:
    bzrlib = None
import os
import os.path
import re
import shutil
import StringIO
import sys

import libbe
import base

if libbe.TESTING == True:
    import doctest
    import unittest


def new():
    return Bzr()

class Bzr(base.VCS):
    """:py:class:`base.VCS` implementation for Bazaar.
    """
    name = 'bzr'
    client = None # bzrlib module

    def __init__(self, *args, **kwargs):
        base.VCS.__init__(self, *args, **kwargs)
        self.versioned = True

    def _vcs_version(self):
        if bzrlib == None:
            return None
        return bzrlib.__version__

    def _vcs_get_user_id(self):
        # excerpted from bzrlib.builtins.cmd_whoami.run()
        try:
            c = bzrlib.branch.Branch.open_containing(self.repo)[0].get_config()
        except errors.NotBranchError:
            c = bzrlib.config.GlobalConfig()
        return c.username()

    def _vcs_detect(self, path):
        if self._u_search_parent_directories(path, '.bzr') != None :
            return True
        return False

    def _vcs_root(self, path):
        """Find the root of the deepest repository containing path."""
        cmd = bzrlib.builtins.cmd_root()
        cmd.outf = StringIO.StringIO()
        cmd.run(filename=path)
        if self.version_cmp(2,2,0) < 0:
            cmd.cleanup_now()
        return cmd.outf.getvalue().rstrip('\n')

    def _vcs_init(self, path):
        cmd = bzrlib.builtins.cmd_init()
        cmd.outf = StringIO.StringIO()
        cmd.run(location=path)
        if self.version_cmp(2,2,0) < 0:
            cmd.cleanup_now()

    def _vcs_destroy(self):
        vcs_dir = os.path.join(self.repo, '.bzr')
        if os.path.exists(vcs_dir):
            shutil.rmtree(vcs_dir)

    def _vcs_add(self, path):
        path = os.path.join(self.repo, path)
        cmd = bzrlib.builtins.cmd_add()
        cmd.outf = StringIO.StringIO()
        kwargs = {'file_ids_from': self.repo}
        if self.repo == os.path.realpath(os.getcwd()):
            # Work around bzr file locking on Windows.
            # See: https://lists.ubuntu.com/archives/bazaar/2011q1/071705.html
            kwargs.pop('file_ids_from')
        cmd.run(file_list=[path], **kwargs)
        if self.version_cmp(2,2,0) < 0:
            cmd.cleanup_now()

    def _vcs_exists(self, path, revision=None):
        manifest = self._vcs_listdir(
            self.repo, revision=revision, recursive=True)
        if path in manifest:
            return True
        return False

    def _vcs_remove(self, path):
        # --force to also remove unversioned files.
        path = os.path.join(self.repo, path)
        cmd = bzrlib.builtins.cmd_remove()
        cmd.outf = StringIO.StringIO()
        cmd.run(file_list=[path], file_deletion_strategy='force')
        if self.version_cmp(2,2,0) < 0:
            cmd.cleanup_now()

    def _vcs_update(self, path):
        pass

    def _parse_revision_string(self, revision=None):
        if revision == None:
            return revision
        rev_opt = bzrlib.option.Option.OPTIONS['revision']
        try:
            rev_spec = rev_opt.type(revision)
        except bzrlib.errors.NoSuchRevisionSpec:
            raise base.InvalidRevision(revision)
        return rev_spec

    def _vcs_get_file_contents(self, path, revision=None):
        if revision == None:
            return base.VCS._vcs_get_file_contents(self, path, revision)
        path = os.path.join(self.repo, path)
        revision = self._parse_revision_string(revision)
        cmd = bzrlib.builtins.cmd_cat()
        cmd.outf = StringIO.StringIO()
        if self.version_cmp(1,6,0) < 0:
            # old bzrlib cmd_cat uses sys.stdout not self.outf for output.
            stdout = sys.stdout
            sys.stdout = cmd.outf
        try:
            cmd.run(filename=path, revision=revision)
        except bzrlib.errors.BzrCommandError, e:
            if 'not present in revision' in str(e):
                raise base.InvalidPath(path, root=self.repo, revision=revision)
            raise
        finally:
            if self.version_cmp(2,0,0) < 0:
                cmd.outf = sys.stdout
                sys.stdout = stdout
            if self.version_cmp(2,2,0) < 0:
                cmd.cleanup_now()
        return cmd.outf.getvalue()

    def _vcs_path(self, id, revision):
        manifest = self._vcs_listdir(
            self.repo, revision=revision, recursive=True)
        return self._u_find_id_from_manifest(id, manifest, revision=revision)

    def _vcs_isdir(self, path, revision):
        try:
            self._vcs_listdir(path, revision)
        except AttributeError, e:
            if 'children' in str(e):
                return False
            raise
        return True

    def _vcs_listdir(self, path, revision, recursive=False):
        path = os.path.join(self.repo, path)
        revision = self._parse_revision_string(revision)
        cmd = bzrlib.builtins.cmd_ls()
        cmd.outf = StringIO.StringIO()
        try:
            if self.version_cmp(2,0,0) >= 0:
                cmd.run(revision=revision, path=path, recursive=recursive)
            else:
                # Pre-2.0 Bazaar (non_recursive)
                # + working around broken non_recursive+path implementation
                #   (https://bugs.launchpad.net/bzr/+bug/158690)
                cmd.run(revision=revision, path=path,
                        non_recursive=False)
        except bzrlib.errors.BzrCommandError, e:
            if 'not present in revision' in str(e):
                raise base.InvalidPath(path, root=self.repo, revision=revision)
            raise
        finally:
            if self.version_cmp(2,2,0) < 0:
                cmd.cleanup_now()
        children = cmd.outf.getvalue().rstrip('\n').splitlines()
        children = [self._u_rel_path(c, path) for c in children]
        if self.version_cmp(2,0,0) < 0 and recursive == False:
            children = [c for c in children if os.path.sep not in c]
        return children

    def _vcs_commit(self, commitfile, allow_empty=False):
        cmd = bzrlib.builtins.cmd_commit()
        cmd.outf = StringIO.StringIO()
        cwd = os.getcwd()
        os.chdir(self.repo)
        try:
            cmd.run(file=commitfile, unchanged=allow_empty)
        except bzrlib.errors.BzrCommandError, e:
            strings = ['no changes to commit.', # bzr 1.3.1
                       'No changes to commit.'] # bzr 1.15.1
            if self._u_any_in_string(strings, str(e)) == True:
                raise base.EmptyCommit()
            raise
        finally:
            os.chdir(cwd)
            if self.version_cmp(2,2,0) < 0:
                cmd.cleanup_now()
        return self._vcs_revision_id(-1)

    def _vcs_revision_id(self, index):
        cmd = bzrlib.builtins.cmd_revno()
        cmd.outf = StringIO.StringIO()
        cmd.run(location=self.repo)
        if self.version_cmp(2,2,0) < 0:
            cmd.cleanup_now()
        current_revision = int(cmd.outf.getvalue())
        if index > current_revision or index < -current_revision:
            return None
        if index >= 0:
            return str(index) # bzr commit 0 is the empty tree.
        return str(current_revision+index+1)

    def _diff(self, revision):
        revision = self._parse_revision_string(revision)
        cmd = bzrlib.builtins.cmd_diff()
        cmd.outf = StringIO.StringIO()
        # for some reason, cmd_diff uses sys.stdout not self.outf for output.
        stdout = sys.stdout
        sys.stdout = cmd.outf
        try:
            status = cmd.run(revision=revision, file_list=[self.repo])
        finally:
            sys.stdout = stdout
            if self.version_cmp(2,2,0) < 0:
                cmd.cleanup_now()
        assert status in [0,1], "Invalid status %d" % status
        return cmd.outf.getvalue()

    def _parse_diff(self, diff_text):
        """_parse_diff(diff_text) -> (new,modified,removed)

        `new`, `modified`, and `removed` are lists of files.

        Example diff text::

          === modified file 'dir/changed'
          --- dir/changed	2010-01-16 01:54:53 +0000
          +++ dir/changed	2010-01-16 01:54:54 +0000
          @@ -1,3 +1,3 @@
           hi
          -there
          +everyone and
           joe
          
          === removed file 'dir/deleted'
          --- dir/deleted	2010-01-16 01:54:53 +0000
          +++ dir/deleted	1970-01-01 00:00:00 +0000
          @@ -1,3 +0,0 @@
          -in
          -the
          -beginning
          
          === removed file 'dir/moved'
          --- dir/moved	2010-01-16 01:54:53 +0000
          +++ dir/moved	1970-01-01 00:00:00 +0000
          @@ -1,4 +0,0 @@
          -the
          -ants
          -go
          -marching
          
          === added file 'dir/moved2'
          --- dir/moved2	1970-01-01 00:00:00 +0000
          +++ dir/moved2	2010-01-16 01:54:34 +0000
          @@ -0,0 +1,4 @@
          +the
          +ants
          +go
          +marching
          
          === added file 'dir/new'
          --- dir/new	1970-01-01 00:00:00 +0000
          +++ dir/new	2010-01-16 01:54:54 +0000
          @@ -0,0 +1,2 @@
          +hello
          +world
          
        """
        new = []
        modified = []
        removed = []
        for line in diff_text.splitlines():
            if not line.startswith('=== '):
                continue
            fields = line.split()
            action = fields[1]
            file = fields[-1].strip("'")
            if action == 'added':
                new.append(file)
            elif action == 'modified':
                modified.append(file)
            elif action == 'removed':
                removed.append(file)
        return (new,modified,removed)

    def _vcs_changed(self, revision):
        return self._parse_diff(self._diff(revision))


if libbe.TESTING == True:
    base.make_vcs_testcase_subclasses(Bzr, sys.modules[__name__])

    unitsuite =unittest.TestLoader().loadTestsFromModule(sys.modules[__name__])
    suite = unittest.TestSuite([unitsuite, doctest.DocTestSuite()])



# Copyright (C) 2005-2012 Aaron Bentley <abentley@panoramicfeedback.com>
#                         Alexander Belchenko <bialix@ukr.net>
#                         Ben Finney <benf@cybersource.com.au>
#                         Chris Ball <cjb@laptop.org>
#                         Gianluca Montecchi <gian@grys.it>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

"""Define the base :py:class:`VCS` (Version Control System) class, which
should be subclassed by other Version Control System backends.  The
base class implements a "do not version" VCS.
"""

import codecs
import os
import os.path
import re
import shutil
import sys
import tempfile
import types

import libbe
import libbe.storage
import libbe.storage.base
import libbe.util.encoding
from libbe.storage.base import EmptyCommit, InvalidRevision, InvalidID
from libbe.util.utility import Dir, search_parent_directories
from libbe.util.subproc import CommandError, invoke
from libbe.util.plugin import import_by_name
import libbe.storage.util.upgrade as upgrade

if libbe.TESTING == True:
    import unittest
    import doctest

    import libbe.ui.util.user

VCS_ORDER = ['arch', 'bzr', 'darcs', 'git', 'hg', 'monotone']
"""List VCS modules in order of preference.

Don't list this module, it is implicitly last.
"""

def set_preferred_vcs(name):
    """Manipulate :py:data:`VCS_ORDER` to place `name` first.

    This is primarily indended for testing purposes.
    """
    global VCS_ORDER
    assert name in VCS_ORDER, \
        'unrecognized VCS %s not in\n  %s' % (name, VCS_ORDER)
    VCS_ORDER.remove(name)
    VCS_ORDER.insert(0, name)

def _get_matching_vcs(matchfn):
    """Return the first module for which matchfn(VCS_instance) is True.

    Searches in :py:data:`VCS_ORDER`.
    """
    for submodname in VCS_ORDER:
        module = import_by_name('libbe.storage.vcs.%s' % submodname)
        vcs = module.new()
        if matchfn(vcs) == True:
            return vcs
    return VCS()

def vcs_by_name(vcs_name):
    """Return the module for the VCS with the given name.

    Searches in :py:data:`VCS_ORDER`.
    """
    if vcs_name == VCS.name:
        return new()
    return _get_matching_vcs(lambda vcs: vcs.name == vcs_name)

def detect_vcs(dir):
    """Return an VCS instance for the vcs being used in this directory.

    Searches in :py:data:`VCS_ORDER`.
    """
    return _get_matching_vcs(lambda vcs: vcs._detect(dir))

def installed_vcs():
    """Return an instance of an installed VCS.

    Searches in :py:data:`VCS_ORDER`.
    """
    return _get_matching_vcs(lambda vcs: vcs.installed())


class VCSNotRooted (libbe.storage.base.ConnectionError):
    def __init__(self, vcs):
        msg = 'VCS not rooted'
        libbe.storage.base.ConnectionError.__init__(self, msg)
        self.vcs = vcs

class VCSUnableToRoot (libbe.storage.base.ConnectionError):
    def __init__(self, vcs):
        msg = 'VCS unable to root'
        libbe.storage.base.ConnectionError.__init__(self, msg)
        self.vcs = vcs

class InvalidPath (InvalidID):
    def __init__(self, path, root, msg=None, **kwargs):
        if msg == None:
            msg = 'Path "%s" not in root "%s"' % (path, root)
        InvalidID.__init__(self, msg=msg, **kwargs)
        self.path = path
        self.root = root

class SpacerCollision (InvalidPath):
    def __init__(self, path, spacer):
        msg = 'Path "%s" collides with spacer directory "%s"' % (path, spacer)
        InvalidPath.__init__(self, path, root=None, msg=msg)
        self.spacer = spacer

class NoSuchFile (InvalidID):
    def __init__(self, pathname, root='.'):
        path = os.path.abspath(os.path.join(root, pathname))
        InvalidID.__init__(self, 'No such file: %s' % path)


class CachedPathID (object):
    """Cache Storage ID <-> path policy.
 
    Paths generated following::

       .../.be/BUGDIR/bugs/BUG/comments/COMMENT
          ^-- root path

    See :py:mod:`libbe.util.id` for a discussion of ID formats.

    Examples
    --------

    >>> dir = Dir()
    >>> os.mkdir(os.path.join(dir.path, '.be'))
    >>> os.mkdir(os.path.join(dir.path, '.be', 'abc'))
    >>> os.mkdir(os.path.join(dir.path, '.be', 'abc', 'bugs'))
    >>> os.mkdir(os.path.join(dir.path, '.be', 'abc', 'bugs', '123'))
    >>> os.mkdir(os.path.join(dir.path, '.be', 'abc', 'bugs', '123', 'comments'))
    >>> os.mkdir(os.path.join(dir.path, '.be', 'abc', 'bugs', '123', 'comments', 'def'))
    >>> os.mkdir(os.path.join(dir.path, '.be', 'abc', 'bugs', '456'))
    >>> open(os.path.join(dir.path, '.be', 'abc', 'values'),
    ...      'w').close()
    >>> open(os.path.join(dir.path, '.be', 'abc', 'bugs', '123', 'values'),
    ...      'w').close()
    >>> open(os.path.join(dir.path, '.be', 'abc', 'bugs', '123', 'comments', 'def', 'values'),
    ...      'w').close()
    >>> c = CachedPathID()
    >>> c.root(dir.path)
    >>> c.id(os.path.join(dir.path, '.be', 'abc', 'bugs', '123', 'comments', 'def', 'values'))
    'def/values'
    >>> c.init()
    >>> sorted(os.listdir(os.path.join(c._root, '.be')))
    ['abc', 'id-cache']
    >>> c.connect()
    >>> c.path('123/values') # doctest: +ELLIPSIS
    u'.../.be/abc/bugs/123/values'
    >>> c.disconnect()
    >>> c.destroy()
    >>> sorted(os.listdir(os.path.join(c._root, '.be')))
    ['abc']
    >>> c.connect() # demonstrate auto init
    >>> sorted(os.listdir(os.path.join(c._root, '.be')))
    ['abc', 'id-cache']
    >>> c.add_id(u'xyz', parent=None) # doctest: +ELLIPSIS
    u'.../.be/xyz'
    >>> c.add_id('xyz/def', parent='xyz') # doctest: +ELLIPSIS
    u'.../.be/xyz/def'
    >>> c.add_id('qrs', parent='123') # doctest: +ELLIPSIS
    u'.../.be/abc/bugs/123/comments/qrs'
    >>> c.disconnect()
    >>> c.connect()
    >>> c.path('qrs') # doctest: +ELLIPSIS
    u'.../.be/abc/bugs/123/comments/qrs'
    >>> c.remove_id('qrs')
    >>> c.path('qrs')
    Traceback (most recent call last):
      ...
    InvalidID: qrs in revision None
    >>> c.disconnect()
    >>> c.destroy()
    >>> dir.cleanup()
    """
    def __init__(self, encoding=None):
        self.encoding = libbe.util.encoding.get_text_file_encoding()
        self._spacer_dirs = ['.be', 'bugs', 'comments']

    def root(self, path):
        self._root = os.path.abspath(path).rstrip(os.path.sep)
        self._cache_path = os.path.join(
            self._root, self._spacer_dirs[0], 'id-cache')

    def init(self, verbose=True, cache=None):
        """Create cache file for an existing .be directory.

        The file contains multiple lines of the form::

            UUID\tPATH
        """
        if cache == None:
            self._cache = {}
        else:
            self._cache = cache
        spaced_root = os.path.join(self._root, self._spacer_dirs[0])
        for dirpath, dirnames, filenames in os.walk(spaced_root,
                                                    followlinks=True):
            if dirpath == spaced_root:
                continue
            try:
                id = self.id(dirpath)
                relpath = dirpath[len(self._root + os.path.sep):]
                if id.count('/') == 0:
                    if verbose == True and id in self._cache:
                        print >> sys.stderr, 'Multiple paths for %s: \n  %s\n  %s' % (id, self._cache[id], relpath)
                    self._cache[id] = relpath
            except InvalidPath:
                pass
        if self._cache != cache:
            self._changed = True
        if cache == None:
            self.disconnect()

    def destroy(self):
        if os.path.exists(self._cache_path):
            os.remove(self._cache_path)

    def connect(self):
        if not os.path.exists(self._cache_path):
            try:
                self.init()
            except IOError:
                raise libbe.storage.base.ConnectionError
        self._cache = {} # key: uuid, value: path
        self._changed = False
        f = codecs.open(self._cache_path, 'r', self.encoding)
        for line in f:
            fields = line.rstrip('\n').split('\t')
            self._cache[fields[0]] = fields[1]
        f.close()

    def disconnect(self):
        if self._changed == True:
            f = codecs.open(self._cache_path, 'w', self.encoding)
            for uuid,path in self._cache.items():
                f.write('%s\t%s\n' % (uuid, path))
            f.close()
        self._cache = {}

    def path(self, id, relpath=False):
        fields = id.split('/', 1)
        uuid = fields[0]
        if len(fields) == 1:
            extra = []
        else:
            extra = fields[1:]
        if uuid not in self._cache:
            self.init(verbose=False, cache=self._cache)
            if uuid not in self._cache:
                raise InvalidID(uuid)
        if relpath == True:
            return os.path.join(self._cache[uuid], *extra)
        return os.path.join(self._root, self._cache[uuid], *extra)

    def add_id(self, id, parent=None):
        if id.count('/') > 0:
            # not a UUID-level path
            assert id.startswith(parent), \
                'Strange ID: "%s" should start with "%s"' % (id, parent)
            path = self.path(id)
        elif id in self._cache:
            # already added
            path = self.path(id)
        else:
            if parent == None:
                parent_path = ''
                spacer = self._spacer_dirs[0]
            else:
                assert parent.count('/') == 0, \
                    'Strange parent ID: "%s" should be UUID' % parent
                parent_path = self.path(parent, relpath=True)
                parent_spacer = parent_path.split(os.path.sep)[-2]
                i = self._spacer_dirs.index(parent_spacer)
                spacer = self._spacer_dirs[i+1]
            path = os.path.join(parent_path, spacer, id)
            self._cache[id] = path
            self._changed = True
            path = os.path.join(self._root, path)
        return path

    def remove_id(self, id):
        if id.count('/') > 0:
            return # not a UUID-level path
        self._cache.pop(id)
        self._changed = True

    def id(self, path):
        path = os.path.join(self._root, path)
        if not path.startswith(self._root + os.path.sep):
            raise InvalidPath(path, self._root)
        path = path[len(self._root + os.path.sep):]
        orig_path = path
        if not path.startswith(self._spacer_dirs[0] + os.path.sep):
            raise InvalidPath(path, self._spacer_dirs[0])
        for spacer in self._spacer_dirs:
            if not path.startswith(spacer + os.path.sep):
                break
            id = path[len(spacer + os.path.sep):]
            fields = path[len(spacer + os.path.sep):].split(os.path.sep,1)
            if len(fields) == 1:
                break
            path = fields[1]
        for spacer in self._spacer_dirs:
            if id.endswith(os.path.sep + spacer):
                raise SpacerCollision(orig_path, spacer)
        if os.path.sep != '/':
            id = id.replace(os.path.sep, '/')
        return id


def new():
    return VCS()

class VCS (libbe.storage.base.VersionedStorage):
    """Implement a 'no-VCS' interface.

    Support for other VCSs can be added by subclassing this class, and
    overriding methods `_vcs_*()` with code appropriate for your VCS.

    The methods `_u_*()` are utility methods available to the `_vcs_*()`
    methods.
    """
    name = 'None'
    client = 'false' # command-line tool for _u_invoke_client

    def __init__(self, *args, **kwargs):
        if 'encoding' not in kwargs:
            kwargs['encoding'] = libbe.util.encoding.get_text_file_encoding()
        libbe.storage.base.VersionedStorage.__init__(self, *args, **kwargs)
        self.versioned = False
        self.interspersed_vcs_files = False
        self.verbose_invoke = False
        self._cached_path_id = CachedPathID()
        self._rooted = False

    def _vcs_version(self):
        """
        Return the VCS version string.
        """
        return '0'

    def _vcs_get_user_id(self):
        """
        Get the VCS's suggested user id (e.g. "John Doe <jdoe@example.com>").
        If the VCS has not been configured with a username, return None.
        """
        return None

    def _vcs_detect(self, path=None):
        """
        Detect whether a directory is revision controlled with this VCS.
        """
        return True

    def _vcs_root(self, path):
        """
        Get the VCS root.  This is the default working directory for
        future invocations.  You would normally set this to the root
        directory for your VCS.
        """
        if os.path.isdir(path) == False:
            path = os.path.dirname(path)
            if path == '':
                path = os.path.abspath('.')
        return path

    def _vcs_init(self, path):
        """
        Begin versioning the tree based at path.
        """
        pass

    def _vcs_destroy(self):
        """
        Remove any files used in versioning (e.g. whatever _vcs_init()
        created).
        """
        pass

    def _vcs_add(self, path):
        """
        Add the already created file at path to version control.
        """
        pass

    def _vcs_exists(self, path, revision=None):
        """
        Does the path exist in a given revision? (True/False)
        """
        raise NotImplementedError('Lazy BE developers')

    def _vcs_remove(self, path):
        """
        Remove the file at path from version control.  Optionally
        remove the file from the filesystem as well.
        """
        pass

    def _vcs_update(self, path):
        """
        Notify the versioning system of changes to the versioned file
        at path.
        """
        pass

    def _vcs_is_versioned(self, path):
        """
        Return true if a path is under version control, False
        otherwise.  You only need to set this if the VCS goes about
        dumping VCS-specific files into the .be directory.

        If you do need to implement this method (e.g. Arch), set
          self.interspersed_vcs_files = True
        """
        assert self.interspersed_vcs_files == False
        raise NotImplementedError

    def _vcs_get_file_contents(self, path, revision=None):
        """
        Get the file contents as they were in a given revision.
        Revision==None specifies the current revision.
        """
        if revision != None:
            raise libbe.storage.base.InvalidRevision(
                'The %s VCS does not support revision specifiers' % self.name)
        path = os.path.join(self.repo, path)
        if not os.path.exists(path):
            return libbe.util.InvalidObject
        if os.path.isdir(path):
            return libbe.storage.base.InvalidDirectory
        f = open(path, 'rb')
        contents = f.read()
        f.close()
        return contents

    def _vcs_path(self, id, revision):
        """
        Return the relative path to object id as of revision.
        
        Revision will not be None.
        """
        raise NotImplementedError

    def _vcs_isdir(self, path, revision):
        """
        Return True if path (as returned by _vcs_path) was a directory
        as of revision, False otherwise.
        
        Revision will not be None.
        """
        raise NotImplementedError

    def _vcs_listdir(self, path, revision):
        """
        Return a list of the contents of the directory path (as
        returned by _vcs_path) as of revision.
        
        Revision will not be None, and ._vcs_isdir(path, revision)
        will be True.
        """
        raise NotImplementedError

    def _vcs_commit(self, commitfile, allow_empty=False):
        """
        Commit the current working directory, using the contents of
        commitfile as the comment.  Return the name of the old
        revision (or None if commits are not supported).

        If allow_empty == False, raise EmptyCommit if there are no
        changes to commit.
        """
        return None

    def _vcs_revision_id(self, index):
        """
        Return the name of the <index>th revision.  Index will be an
        integer (possibly <= 0).  The choice of which branch to follow
        when crossing branches/merges is not defined.

        Return None if revision IDs are not supported, or if the
        specified revision does not exist.
        """
        return None

    def _vcs_changed(self, revision):
        """
        Return a tuple of lists of ids
          (new, modified, removed)
        from the specified revision to the current situation.
        """
        return ([], [], [])

    def version(self):
        # Cache version string for efficiency.
        if not hasattr(self, '_version'):
            self._version = self._vcs_version()
        return self._version

    def version_cmp(self, *args):
        """Compare the installed VCS version `V_i` with another version
        `V_o` (given in `*args`).  Returns

           === ===============
            1  if `V_i > V_o`
            0  if `V_i == V_o`
           -1  if `V_i < V_o`
           === ===============

        Examples
        --------

        >>> v = VCS(repo='.')
        >>> v._version = '2.3.1 (release)'
        >>> v.version_cmp(2,3,1)
        0
        >>> v.version_cmp(2,3,2)
        -1
        >>> v.version_cmp(2,3,'a',5)
        1
        >>> v.version_cmp(2,3,0)
        1
        >>> v.version_cmp(2,3,1,'a',5)
        1
        >>> v.version_cmp(2,3,1,1)
        -1
        >>> v.version_cmp(3)
        -1
        >>> v._version = '2.0.0pre2'
        >>> v._parsed_version = None
        >>> v.version_cmp(3)
        -1
        >>> v.version_cmp(2,0,1)
        -1
        >>> v.version_cmp(2,0,0,'pre',1)
        1
        >>> v.version_cmp(2,0,0,'pre',2)
        0
        >>> v.version_cmp(2,0,0,'pre',3)
        -1
        >>> v.version_cmp(2,0,0,'a',3)
        1
        >>> v.version_cmp(2,0,0,'rc',1)
        -1
        """
        if not hasattr(self, '_parsed_version') \
                or self._parsed_version == None:
            num_part = self.version().split(' ')[0]
            self._parsed_version = []
            for num in num_part.split('.'):
                try:
                    self._parsed_version.append(int(num))
                except ValueError, e:
                    # bzr version number might contain non-numerical tags
                    splitter = re.compile(r'[\D]') # Match non-digits
                    splits = splitter.split(num)
                    # if len(tag) > 1 some splits will be empty; remove
                    splits = filter(lambda s: s != '', splits)
                    tag_starti = len(splits[0])
                    num_starti = num.find(splits[1], tag_starti)
                    tag = num[tag_starti:num_starti]
                    self._parsed_version.append(int(splits[0]))
                    self._parsed_version.append(tag)
                    self._parsed_version.append(int(splits[1]))
        for current,other in zip(self._parsed_version, args):
            if type(current) != type (other):
                # one of them is a pre-release string
                if type(current) != types.IntType:
                    return -1
                else:
                    return 1
            c = cmp(current,other)
            if c != 0:
                return c
        # see if one is longer than the other
        verlen = len(self._parsed_version)
        arglen = len(args)
        if verlen == arglen:
            return 0
        elif verlen > arglen:
            if type(self._parsed_version[arglen]) != types.IntType:
                return -1 # self is a prerelease
            else:
                return 1
        else:
            if type(args[verlen]) != types.IntType:
                return 1 # args is a prerelease
            else:
                return -1

    def installed(self):
        if self.version() != None:
            return True
        return False

    def get_user_id(self):
        """
        Get the VCS's suggested user id (e.g. "John Doe <jdoe@example.com>").
        If the VCS has not been configured with a username, return None.
        You can override the automatic lookup procedure by setting the
        VCS.user_id attribute to a string of your choice.
        """
        if not hasattr(self, 'user_id'):
            self.user_id = self._vcs_get_user_id()
            if self.user_id == None:
                # guess missing info
                name = libbe.ui.util.user.get_fallback_fullname()
                email = libbe.ui.util.user.get_fallback_email()
                self.user_id = libbe.ui.util.user.create_user_id(name, email)
        return self.user_id

    def _detect(self, path='.'):
        """
        Detect whether a directory is revision controlled with this VCS.
        """
        return self._vcs_detect(path)

    def root(self):
        """Set the root directory to the path's VCS root.

        This is the default working directory for future invocations.
        Consider the following usage case:

        You have a project rooted in::

          /path/to/source/

        by which I mean the VCS repository is in, for example::

          /path/to/source/.bzr

        However, you're of in some subdirectory like::

          /path/to/source/ui/testing

        and you want to comment on a bug.  `root` will locate your VCS
        root (``/path/to/source/``) and set the repo there.  This
        means that it doesn't matter where you are in your project
        tree when you call "be COMMAND", it always acts as if you called
        it from the VCS root.
        """
        if self._detect(self.repo) == False:
            raise VCSUnableToRoot(self)
        root = self._vcs_root(self.repo)
        self.repo = os.path.realpath(root)
        if os.path.isdir(self.repo) == False:
            self.repo = os.path.dirname(self.repo)
        self.be_dir = os.path.join(
            self.repo, self._cached_path_id._spacer_dirs[0])
        self._cached_path_id.root(self.repo)
        self._rooted = True

    def _init(self):
        """
        Begin versioning the tree based at self.repo.
        Also roots the vcs at path.

        See Also
        --------
        root : called if the VCS has already been initialized.
        """
        if not os.path.exists(self.repo) or not os.path.isdir(self.repo):
            raise VCSUnableToRoot(self)
        if self._vcs_detect(self.repo) == False:
            self._vcs_init(self.repo)
        if self._rooted == False:
            self.root()
        os.mkdir(self.be_dir)
        self._vcs_add(self._u_rel_path(self.be_dir))
        self._setup_storage_version()
        self._cached_path_id.init()

    def _destroy(self):
        self._vcs_destroy()
        self._cached_path_id.destroy()
        if os.path.exists(self.be_dir):
            shutil.rmtree(self.be_dir)

    def _connect(self):
        if self._rooted == False:
            self.root()
        if not os.path.isdir(self.be_dir):
            raise libbe.storage.base.ConnectionError(self)
        self._cached_path_id.connect()
        self.check_storage_version()

    def _disconnect(self):
        self._cached_path_id.disconnect()

    def path(self, id, revision=None, relpath=True):
        if revision == None:
            path = self._cached_path_id.path(id)
            if relpath == True:
                return self._u_rel_path(path)
            return path
        path = self._vcs_path(id, revision)
        if relpath == True:
            return path
        return os.path.join(self.repo, path)

    def _add_path(self, path, directory=False):
        relpath = self._u_rel_path(path)
        reldirs = relpath.split(os.path.sep)
        if directory == False:
            reldirs = reldirs[:-1]
        dir = self.repo
        for reldir in reldirs:
            dir = os.path.join(dir, reldir)
            if not os.path.exists(dir):
                os.mkdir(dir)
                self._vcs_add(self._u_rel_path(dir))
            elif not os.path.isdir(dir):
                raise libbe.storage.base.InvalidDirectory
        if directory == False:
            if not os.path.exists(path):
                open(path, 'w').close()
            self._vcs_add(self._u_rel_path(path))

    def _add(self, id, parent=None, **kwargs):
        path = self._cached_path_id.add_id(id, parent)
        self._add_path(path, **kwargs)

    def _exists(self, id, revision=None):
        if revision == None:
            try:
                path = self.path(id, revision, relpath=False)
            except InvalidID, e:
                return False
            return os.path.exists(path)
        path = self.path(id, revision, relpath=True)
        return self._vcs_exists(relpath, revision)

    def _remove(self, id):
        path = self._cached_path_id.path(id)
        if os.path.exists(path):
            if os.path.isdir(path) and len(self.children(id)) > 0:
                raise libbe.storage.base.DirectoryNotEmpty(id)
            self._vcs_remove(self._u_rel_path(path))
            if os.path.exists(path):
                if os.path.isdir(path):
                    os.rmdir(path)
                else:
                    os.remove(path)
        self._cached_path_id.remove_id(id)

    def _recursive_remove(self, id):
        path = self._cached_path_id.path(id)
        for dirpath,dirnames,filenames in os.walk(path, topdown=False):
            filenames.extend(dirnames)
            for f in filenames:
                fullpath = os.path.join(dirpath, f)
                if os.path.exists(fullpath) == False:
                    continue
                self._vcs_remove(self._u_rel_path(fullpath))
        if os.path.exists(path):
            shutil.rmtree(path)
        path = self._cached_path_id.path(id, relpath=True)
        for id,p in self._cached_path_id._cache.items():
            if p.startswith(path):
                self._cached_path_id.remove_id(id)

    def _ancestors(self, id=None, revision=None):
        if id==None:
            path = self.be_dir
        else:
            path = self.path(id, revision, relpath=False)
        ancestors = []
        while True:
            if not path.startswith(self.repo + os.path.sep):
                break
            path = os.path.dirname(path)
            try:
                id = self._u_path_to_id(path)
                ancestors.append(id)
            except (SpacerCollision, InvalidPath):
                pass    
        return ancestors

    def _children(self, id=None, revision=None):
        if revision == None:
            isdir = os.path.isdir
            listdir = os.listdir
        else:
            isdir = lambda path : self._vcs_isdir(
                self._u_rel_path(path), revision)
            listdir = lambda path : self._vcs_listdir(
                self._u_rel_path(path), revision)
        if id==None:
            path = self.be_dir
        else:
            path = self.path(id, revision, relpath=False)
        if isdir(path) == False: 
            return []
        children = listdir(path)
        for i,c in enumerate(children):
            if c in self._cached_path_id._spacer_dirs:
                children[i] = None
                children.extend([os.path.join(c, c2) for c2 in
                                 listdir(os.path.join(path, c))])
            elif c in ['id-cache', 'version']:
                children[i] = None
            elif self.interspersed_vcs_files \
                    and self._vcs_is_versioned(c) == False:
                children[i] = None
        for i,c in enumerate(children):
            if c == None: continue
            cpath = os.path.join(path, c)
            if self.interspersed_vcs_files == True \
                    and revision != None \
                    and self._vcs_is_versioned(cpath) == False:
                children[i] = None
            else:
                children[i] = self._u_path_to_id(cpath)
        return [c for c in children if c != None]

    def _get(self, id, default=libbe.util.InvalidObject, revision=None):
        try:
            relpath = self.path(id, revision, relpath=True)
            contents = self._vcs_get_file_contents(relpath, revision)
        except InvalidID, e:
            if default == libbe.util.InvalidObject:
                raise e
            return default
        if contents in [libbe.storage.base.InvalidDirectory,
                        libbe.util.InvalidObject] \
                or len(contents) == 0:
            if default == libbe.util.InvalidObject:
                raise InvalidID(id, revision)
            return default
        return contents

    def _set(self, id, value):
        try:
            path = self._cached_path_id.path(id)
        except InvalidID, e:
            raise
        if not os.path.exists(path):
            raise InvalidID(id)
        if os.path.isdir(path):
            raise libbe.storage.base.InvalidDirectory(id)
        f = open(path, "wb")
        f.write(value)
        f.close()
        self._vcs_update(self._u_rel_path(path))

    def _commit(self, summary, body=None, allow_empty=False):
        summary = summary.strip()+'\n'
        if body is not None:
            summary += '\n' + body.strip() + '\n'
        descriptor, filename = tempfile.mkstemp()
        revision = None
        try:
            temp_file = os.fdopen(descriptor, 'wb')
            temp_file.write(summary)
            temp_file.flush()
            revision = self._vcs_commit(filename, allow_empty=allow_empty)
            temp_file.close()
        finally:
            os.remove(filename)
        return revision

    def revision_id(self, index=None):
        if index == None:
            return None
        try:
            if int(index) != index:
                raise InvalidRevision(index)
        except ValueError:
            raise InvalidRevision(index)
        revid = self._vcs_revision_id(index)
        if revid == None:
            raise libbe.storage.base.InvalidRevision(index)
        return revid

    def changed(self, revision):
        new,mod,rem = self._vcs_changed(revision)
        def paths_to_ids(paths):
            for p in paths:
                try:
                    id = self._u_path_to_id(p)
                    yield id
                except (SpacerCollision, InvalidPath):
                    pass
        new_id = list(paths_to_ids(new))
        mod_id = list(paths_to_ids(mod))
        rem_id = list(paths_to_ids(rem))
        return (new_id, mod_id, rem_id)

    def _u_any_in_string(self, list, string):
        """Return True if any of the strings in list are in string.
        Otherwise return False.
        """
        for list_string in list:
            if list_string in string:
                return True
        return False

    def _u_invoke(self, *args, **kwargs):
        if 'cwd' not in kwargs:
            kwargs['cwd'] = self.repo
        if 'verbose' not in kwargs:
            kwargs['verbose'] = self.verbose_invoke
        if 'encoding' not in kwargs:
            kwargs['encoding'] = self.encoding
        return invoke(*args, **kwargs)

    def _u_invoke_client(self, *args, **kwargs):
        cl_args = [self.client]
        cl_args.extend(args)
        return self._u_invoke(cl_args, **kwargs)

    def _u_search_parent_directories(self, path, filename):
        """Find the file (or directory) named filename in path or in any of
        path's parents.

        e.g.
          search_parent_directories("/a/b/c", ".be")
        will return the path to the first existing file from
          /a/b/c/.be
          /a/b/.be
          /a/.be
          /.be
        or None if none of those files exist.
        """
        try:
            ret = search_parent_directories(path, filename)
        except AssertionError, e:
            return None
        return ret

    def _u_find_id_from_manifest(self, id, manifest, revision=None):
        """Search for the relative path to id using manifest, a list of all
        files.
        
        Returns None if the id is not found.
        """
        be_dir = self._cached_path_id._spacer_dirs[0]
        be_dir_sep = self._cached_path_id._spacer_dirs[0] + os.path.sep
        files = [f for f in manifest if f.startswith(be_dir_sep)]
        for file in files:
            if not file.startswith(be_dir+os.path.sep):
                continue
            parts = file.split(os.path.sep)
            dir = parts.pop(0) # don't add the first spacer dir
            for part in parts[:-1]:
                dir = os.path.join(dir, part)
                if not dir in files:
                    files.append(dir)
        for file in files:
            try:
                p_id = self._u_path_to_id(file)
                if p_id == id:
                    return file
            except (SpacerCollision, InvalidPath):
                pass
        raise InvalidID(id, revision=revision)

    def _u_find_id(self, id, revision):
        """Search for the relative path to id as of revision.

        Returns None if the id is not found.
        """
        assert self._rooted == True
        be_dir = self._cached_path_id._spacer_dirs[0]
        stack = [(be_dir, be_dir)]
        while len(stack) > 0:
            path,long_id = stack.pop()
            if long_id.endswith('/'+id):
                return path
            if self._vcs_isdir(path, revision) == False:
                continue
            for child in self._vcs_listdir(path, revision):
                stack.append((os.path.join(path, child),
                              '/'.join([long_id, child])))
        raise InvalidID(id, revision=revision)

    def _u_path_to_id(self, path):
        return self._cached_path_id.id(path)

    def _u_rel_path(self, path, root=None):
        """Return the relative path to path from root.

        Examples:

        >>> vcs = new()
        >>> vcs._u_rel_path("/a.b/c/.be", "/a.b/c")
        '.be'
        >>> vcs._u_rel_path("/a.b/c/", "/a.b/c")
        '.'
        >>> vcs._u_rel_path("/a.b/c/", "/a.b/c/")
        '.'
        >>> vcs._u_rel_path("./a", ".")
        'a'
        """
        if root == None:
            if self.repo == None:
                raise VCSNotRooted(self)
            root = self.repo
        path = os.path.abspath(path)
        absRoot = os.path.abspath(root)
        absRootSlashedDir = os.path.join(absRoot,"")
        if path in [absRoot, absRootSlashedDir]:
            return '.'
        if not path.startswith(absRootSlashedDir):
            raise InvalidPath(path, absRootSlashedDir)
        relpath = path[len(absRootSlashedDir):]
        return relpath

    def _u_abspath(self, path, root=None):
        """Return the absolute path from a path relative to root.

        Examples
        --------

        >>> vcs = new()
        >>> vcs._u_abspath(".be", "/a.b/c")
        '/a.b/c/.be'
        """
        if root == None:
            assert self.repo != None, "VCS not rooted"
            root = self.repo
        return os.path.abspath(os.path.join(root, path))

    def _u_parse_commitfile(self, commitfile):
        """Split the commitfile created in self.commit() back into summary and
        header lines.
        """
        f = codecs.open(commitfile, 'r', self.encoding)
        summary = f.readline()
        body = f.read()
        body.lstrip('\n')
        if len(body) == 0:
            body = None
        f.close()
        return (summary, body)

    def check_storage_version(self):
        version = self.storage_version()
        if version != libbe.storage.STORAGE_VERSION:
            upgrade.upgrade(self.repo, version)

    def storage_version(self, revision=None, path=None):
        """Return the storage version of the on-disk files.

        See Also
        --------
        libbe.storage.util.upgrade
        """
        if path == None:
            path = os.path.join(self.repo, '.be', 'version')
        if not os.path.exists(path):
            raise libbe.storage.InvalidStorageVersion(None)
        if revision == None: # don't require connection
            return libbe.util.encoding.get_file_contents(
                path, decode=True).rstrip()
        relpath = self._u_rel_path(path)
        contents = self._vcs_get_file_contents(relpath, revision=revision)
        if type(contents) != types.UnicodeType:
            contents = unicode(contents, self.encoding)
        return contents.strip()

    def _setup_storage_version(self):
        """
        Requires disk access.
        """
        assert self._rooted == True
        path = os.path.join(self.be_dir, 'version')
        if not os.path.exists(path):
            libbe.util.encoding.set_file_contents(path,
                libbe.storage.STORAGE_VERSION+'\n')
            self._vcs_add(self._u_rel_path(path))


if libbe.TESTING == True:
    class VCSTestCase (unittest.TestCase):
        """
        Test cases for base VCS class (in addition to the Storage test
        cases).
        """

        Class = VCS

        def __init__(self, *args, **kwargs):
            super(VCSTestCase, self).__init__(*args, **kwargs)
            self.dirname = None

        def setUp(self):
            """Set up test fixtures for Storage test case."""
            super(VCSTestCase, self).setUp()
            self.dir = Dir()
            self.dirname = self.dir.path
            self.s = self.Class(repo=self.dirname)
            if self.s.installed() == True:
                self.s.init()
                self.s.connect()

        def tearDown(self):
            super(VCSTestCase, self).tearDown()
            if self.s.installed() == True:
                self.s.disconnect()
                self.s.destroy()
            self.dir.cleanup()

    class VCS_installed_TestCase (VCSTestCase):
        def test_installed(self):
            """See if the VCS is installed.
            """
            self.failUnless(self.s.installed() == True,
                            '%(name)s VCS not found' % vars(self.Class))


    class VCS_detection_TestCase (VCSTestCase):
        def test_detection(self):
            """See if the VCS detects its installed repository
            """
            if self.s.installed():
                self.s.disconnect()
                self.failUnless(self.s._detect(self.dirname) == True,
                    'Did not detected %(name)s VCS after initialising'
                    % vars(self.Class))
                self.s.connect()

        def test_no_detection(self):
            """See if the VCS detects its installed repository
            """
            if self.s.installed() and self.Class.name != 'None':
                self.s.disconnect()
                self.s.destroy()
                self.failUnless(self.s._detect(self.dirname) == False,
                    'Detected %(name)s VCS before initialising'
                    % vars(self.Class))
                self.s.init()
                self.s.connect()

        def test_vcs_repo_in_specified_root_path(self):
            """VCS root directory should be in specified root path."""
            rp = os.path.realpath(self.s.repo)
            dp = os.path.realpath(self.dirname)
            vcs_name = self.Class.name
            self.failUnless(
                dp == rp or rp == None,
                "%(vcs_name)s VCS root in wrong dir (%(dp)s %(rp)s)" % vars())

    class VCS_get_user_id_TestCase(VCSTestCase):
        """Test cases for VCS.get_user_id method."""

        def test_get_existing_user_id(self):
            """Should get the existing user ID."""
            if self.s.installed():
                user_id = self.s.get_user_id()
                if user_id == None:
                    return
                name,email = libbe.ui.util.user.parse_user_id(user_id)
                if email != None:
                    self.failUnless('@' in email, email)

    def make_vcs_testcase_subclasses(vcs_class, namespace):
        c = vcs_class()
        if c.installed():
            if c.versioned == True:
                libbe.storage.base.make_versioned_storage_testcase_subclasses(
                    vcs_class, namespace)
            else:
                libbe.storage.base.make_storage_testcase_subclasses(
                    vcs_class, namespace)

        if namespace != sys.modules[__name__]:
            # Make VCSTestCase subclasses for vcs_class in the namespace.
            vcs_testcase_classes = [
                c for c in (
                    ob for ob in globals().values() if isinstance(ob, type))
                if issubclass(c, VCSTestCase) \
                    and c.Class == VCS]

            for base_class in vcs_testcase_classes:
                testcase_class_name = vcs_class.__name__ + base_class.__name__
                testcase_class_bases = (base_class,)
                testcase_class_dict = dict(base_class.__dict__)
                testcase_class_dict['Class'] = vcs_class
                testcase_class = type(
                    testcase_class_name, testcase_class_bases, testcase_class_dict)
                setattr(namespace, testcase_class_name, testcase_class)

    make_vcs_testcase_subclasses(VCS, sys.modules[__name__])

    unitsuite =unittest.TestLoader().loadTestsFromModule(sys.modules[__name__])
    suite = unittest.TestSuite([unitsuite, doctest.DocTestSuite()])



# Copyright (C) 2005-2012 Aaron Bentley <abentley@panoramicfeedback.com>
#                         Chris Ball <cjb@laptop.org>
#                         Gianluca Montecchi <gian@grys.it>
#                         Marien Zwart <marien.zwart@gmail.com>
#                         Robert Lehmann <mail@robertlehmann.de>
#                         Thomas Gerigk <tgerigk@gmx.de>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

import libbe
import libbe.command
import libbe.command.util


class Assign (libbe.command.Command):
    u"""Assign an individual or group to fix a bug

    >>> import sys
    >>> import libbe.bugdir
    >>> bd = libbe.bugdir.SimpleBugDir(memory=False)
    >>> io = libbe.command.StringInputOutput()
    >>> io.stdout = sys.stdout
    >>> ui = libbe.command.UserInterface(io=io)
    >>> ui.storage_callbacks.set_storage(bd.storage)
    >>> cmd = Assign(ui=ui)

    >>> bd.bug_from_uuid('a').assigned is None
    True
    >>> ui._user_id = u'Fran\xe7ois'
    >>> ret = ui.run(cmd, args=['-', '/a'])
    >>> bd.flush_reload()
    >>> bd.bug_from_uuid('a').assigned
    u'Fran\\xe7ois'

    >>> ret = ui.run(cmd, args=['someone', '/a', '/b'])
    >>> bd.flush_reload()
    >>> bd.bug_from_uuid('a').assigned
    'someone'
    >>> bd.bug_from_uuid('b').assigned
    'someone'

    >>> ret = ui.run(cmd, args=['none', '/a'])
    >>> bd.flush_reload()
    >>> bd.bug_from_uuid('a').assigned is None
    True
    >>> ui.cleanup()
    >>> bd.cleanup()
    """
    name = 'assign'

    def __init__(self, *args, **kwargs):
        libbe.command.Command.__init__(self, *args, **kwargs)
        self.args.extend([
                libbe.command.Argument(
                    name='assigned', metavar='ASSIGNED', default=None,
                    completion_callback=libbe.command.util.complete_assigned),
                libbe.command.Argument(
                    name='bug-id', metavar='BUG-ID', default=None,
                    repeatable=True,
                    completion_callback=libbe.command.util.complete_bug_id),
                ])

    def _run(self, **params):
        assigned = parse_assigned(self, params['assigned'])
        bugdirs = self._get_bugdirs()
        for bug_id in params['bug-id']:
            bugdir,bug,comment = (
                libbe.command.util.bugdir_bug_comment_from_user_id(
                    bugdirs, bug_id))
            if bug.assigned != assigned:
                bug.assigned = assigned
                if bug.status == 'open':
                    bug.status = 'assigned'
        return 0

    def _long_help(self):
        return """
Assign a person to fix a bug.

Assigneds should be the person's Bugs Everywhere identity, the same
string that appears in Creator fields.

Special assigned strings:
  "-"      assign the bug to yourself
  "none"   un-assigns the bug
"""

def parse_assigned(command, assigned):
    """Standard processing for the 'assigned' Argument.
    """
    if assigned == 'none':
        assigned = None
    elif assigned == '-':
        assigned = command._get_user_id()
    return assigned



# Copyright (C) 2005-2012 Aaron Bentley <abentley@panoramicfeedback.com>
#                         Chris Ball <cjb@laptop.org>
#                         Gianluca Montecchi <gian@grys.it>
#                         Marien Zwart <marien.zwart@gmail.com>
#                         Thomas Gerigk <tgerigk@gmx.de>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

import libbe
import libbe.command
import libbe.command.util
import libbe.command.depend
import libbe.util.id


class Target (libbe.command.Command):
    """Assorted bug target manipulations and queries

    >>> import os, StringIO, sys
    >>> import libbe.bugdir
    >>> bd = libbe.bugdir.SimpleBugDir(memory=False)
    >>> io = libbe.command.StringInputOutput()
    >>> io.stdout = sys.stdout
    >>> ui = libbe.command.UserInterface(io=io)
    >>> ui.storage_callbacks.set_storage(bd.storage)
    >>> cmd = Target(ui=ui)

    >>> ret = ui.run(cmd, args=['/a'])
    No target assigned.
    >>> ret = ui.run(cmd, args=['/a', 'tomorrow'])
    >>> ret = ui.run(cmd, args=['/a'])
    tomorrow

    >>> ui.io.stdout = StringIO.StringIO()
    >>> ret = ui.run(cmd, {'resolve':True}, ['tomorrow'])
    >>> output = ui.io.get_stdout().strip()
    >>> bd.flush_reload()
    >>> target = bd.bug_from_uuid(output)
    >>> print target.summary
    tomorrow
    >>> print target.severity
    target

    >>> ui.io.stdout = sys.stdout
    >>> ret = ui.run(cmd, args=['/a', 'none'])
    >>> ret = ui.run(cmd, args=['/a'])
    No target assigned.
    >>> ui.cleanup()
    >>> bd.cleanup()
    """
    name = 'target'

    def __init__(self, *args, **kwargs):
        libbe.command.Command.__init__(self, *args, **kwargs)
        self.options.extend([
                libbe.command.Option(name='resolve', short_name='r',
                    help="Print the UUID for the target bug whose summary "
                    "matches TARGET.  If TARGET is not given, print the UUID "
                    "of the current bugdir target."),
                libbe.command.Option(name='bugdir', short_name='b',
                    help='Short bugdir UUID for the target resolution.  You '
                    'only need to set this if you have multiple bugdirs in '
                    'your repository.',
                    arg=libbe.command.Argument(
                        name='bugdir', metavar='ID', default=None,
                        completion_callback=libbe.command.util.complete_bugdir_id)),
                ])
        self.args.extend([
                libbe.command.Argument(
                    name='id', metavar='BUG-ID', optional=True,
                    completion_callback=libbe.command.util.complete_bug_id),
                libbe.command.Argument(
                    name='target', metavar='TARGET', optional=True,
                    completion_callback=complete_target),
                ])

    def _run(self, **params):
        if params['resolve'] == False:
            if params['id'] == None:
                raise libbe.command.UserError('Please specify a bug id.')
        else:
            if params['target'] != None:
                raise libbe.command.UserError('Too many arguments')
            params['target'] = params.pop('id')
        bugdirs = self._get_bugdirs()
        if params['resolve'] == True:
            if params['bugdir']:
                bugdir = bugdirs[bugdir]
            elif len(bugdirs) == 1:
                bugdir = bugdirs.values()[0]
            else:
                raise libbe.command.UserError(
                    'Ambiguous bugdir {}'.format(sorted(bugdirs.values())))
            bug = bug_from_target_summary(bugdirs, bugdir, params['target'])
            if bug == None:
                print >> self.stdout, 'No target assigned.'
            else:
                print >> self.stdout, bug.uuid
            return 0
        bugdir,bug,comment = (
            libbe.command.util.bugdir_bug_comment_from_user_id(
                bugdirs, params['id']))
        if params['target'] == None:
            target = bug_target(bugdirs, bug)
            if target == None:
                print >> self.stdout, 'No target assigned.'
            else:
                print >> self.stdout, target.summary
        else:
            if params['target'] == 'none':
                target = remove_target(bugdirs, bug)
            else:
                target = add_target(bugdirs, bugdir, bug, params['target'])
        return 0

    def usage(self):
        return 'usage: be %(name)s BUG-ID [TARGET]\nor:    be %(name)s --resolve [TARGET]' \
            % vars(self.__class__)

    def _long_help(self):
        return """
Assorted bug target manipulations and queries.

If no target is specified, the bug's current target is printed.  If
TARGET is specified, it will be assigned to the bug, creating a new
target bug if necessary.

Targets are free-form; any text may be specified.  They will generally
be milestone names or release numbers.  The value "none" can be used
to unset the target.

In the alternative `be target --resolve TARGET` form, print the UUID
of the target-bug with summary TARGET.  If target is not given, return
use the bugdir's current target (see `be set`).

If you want to list all bugs blocking the current target, try
  $ be depend --status -closed,fixed,wontfix --severity -target \
    $(be target --resolve)

If you want to set the current bugdir target by summary (rather than
by UUID), try
  $ be set target $(be target --resolve SUMMARY)
"""

def bug_from_target_summary(bugdirs, bugdir, summary=None):
    if summary == None:
        if bugdir.target == None:
            return None
        else:
            return bugdir.bug_from_uuid(bugdir.target)
    matched = []
    for uuid in bugdir.uuids():
        bug = bugdir.bug_from_uuid(uuid)
        if bug.severity == 'target' and bug.summary == summary:
            matched.append(bug)
    if len(matched) == 0:
        return None
    if len(matched) > 1:
        raise Exception('Several targets with same summary:  %s'
                        % '\n  '.join([bug.uuid for bug in matched]))
    return matched[0]

def bug_target(bugdirs, bug):
    if bug.severity == 'target':
        return bug
    matched = []
    for blocked in libbe.command.depend.get_blocks(bugdirs, bug):
        if blocked.severity == 'target':
            matched.append(blocked)
    if len(matched) == 0:
        return None
    if len(matched) > 1:
        raise Exception('This bug (%s) blocks several targets:  %s'
                        % (bug.uuid,
                           '\n  '.join([b.uuid for b in matched])))
    return matched[0]

def remove_target(bugdirs, bug):
    target = bug_target(bugdirs, bug)
    libbe.command.depend.remove_block(target, bug)
    return target

def add_target(bugdirs, bugdir, bug, summary):
    target = bug_from_target_summary(bugdirs, bugdir, summary)
    if target == None:
        target = bugdir.new_bug(summary=summary)
        target.severity = 'target'
    libbe.command.depend.add_block(target, bug)
    return target

def targets(bugdirs):
    """Generate all possible target bug summaries."""
    for bugdir in bugdirs.values():
        bugdir.load_all_bugs()
        for bug in bugdir:
            if bug.severity == 'target':
                yield bug.summary

def target_dict(bugdirs):
    """
    Return a dict with bug UUID keys and bug summary values for all
    target bugs.
    """
    ret = {}
    for bug in targets(bugdirs):
        ret[bug.uuid] = bug
    return ret

def complete_target(command, argument, fragment=None):
    """List possible command completions for fragment."""
    return targets(command._get_bugdirs())



# Copyright (C) 2010-2012 Chris Ball <cjb@laptop.org>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

"""Define the :py:class:`ServeCommands` serving BE Commands over HTTP.

See Also
--------
:py:meth:`be-libbe.command.base.Command._run_remote` : the associated client
"""

import logging
import os.path
import posixpath
import re
import urllib
import wsgiref.simple_server

import libbe
import libbe.command
import libbe.command.base
import libbe.storage.util.mapfile
import libbe.util.wsgi
import libbe.version

if libbe.TESTING:
    import copy
    import doctest
    import StringIO
    import sys
    import unittest
    import wsgiref.validate
    try:
        import cherrypy.test.webtest
        cherrypy_test_webtest = True
    except ImportError:
        cherrypy_test_webtest = None

    import libbe.bugdir
    import libbe.command.list

    
class ServerApp (libbe.util.wsgi.WSGI_AppObject,
                 libbe.util.wsgi.WSGI_DataObject):
    """WSGI server for a BE Command invocation over HTTP.

    RESTful_ WSGI request handler for serving the
    libbe.command.base.Command._run_remote backend with GET, POST, and
    HEAD commands.

    This serves all commands from a single, persistant storage
    instance, usually a VCS-based repository located on the local
    machine.
    """
    server_version = "BE-command-server/" + libbe.version.version()

    def __init__(self, storage=None, notify=False, **kwargs):
        super(ServerApp, self).__init__(
            urls=[
                (r'^run/?$', self.run),
                ],
            **kwargs)
        self.storage = storage
        self.ui = libbe.command.base.UserInterface()
        self.notify = notify

    # handlers
    def run(self, environ, start_response):
        self.check_login(environ)
        data = self.post_data(environ)
        source = 'post'
        try:
            name = data['command']
        except KeyError:
            raise libbe.util.wsgi.HandlerError(
                libbe.util.http.HTTP_USER_ERROR, 'UnknownCommand')
        parameters = data.get('parameters', {})
        try:
            Class = libbe.command.get_command_class(command_name=name)
        except libbe.command.UnknownCommand, e:
            raise libbe.util.wsgi.HandlerError(
                libbe.util.http.HTTP_USER_ERROR,
                'UnknownCommand {0}'.format(e))
        command = Class(ui=self.ui)
        self.ui.setup_command(command)
        arguments = [option.arg for option in command.options
                     if option.arg is not None]
        arguments.extend(command.args)
        for argument in arguments:
            if argument.name not in parameters:
                parameters[argument.name] = argument.default
        command.status = command._run(**parameters)  # already parsed params
        assert command.status == 0, command.status
        stdout = self.ui.io.get_stdout()
        if self.notify:  # TODO, check what notify does
            self._notify(environ, 'run', command)
        return self.ok_response(environ, start_response, stdout)

    # handler utility functions
    def _parse_post(self, post):
        return libbe.storage.util.mapfile.parse(post)

    def check_login(self, environ):
        user = environ.get('be-auth.user', None)
        if user is not None:  # we're running under AuthenticationApp
            if environ['REQUEST_METHOD'] == 'POST':
                # TODO: better detection of commands requiring writes
                if user == 'guest' or self.storage.is_writeable() == False:
                    raise _Unauthorized() # only non-guests allowed to write
            # allow read-only commands for all users

    def _notify(self, environ, command, id, params):
        message = self._format_notification(environ, command, id, params)
        self._submit_notification(message)

    def _format_notification(self, environ, command, id, params):
        key_length = len('command')
        for key,value in params:
            if len(key) > key_length and '\n' not in str(value):
                key_length = len(key)
        key_length += 1
        lines = []
        multi_line_params = []
        for key,value in [('address', environ.get('REMOTE_ADDR', '-')),
                          ('command', command), ('id', id)]+params:
            v = str(value)
            if '\n' in v:
                multi_line_params.append((key,v))
                continue
            lines.append('%*.*s %s' % (key_length, key_length, key+':', v))
        lines.append('')
        for key,value in multi_line_params:
            lines.extend(['=== START %s ===' % key, v,
                          '=== STOP %s ===' % key, ''])
        lines.append('')
        return '\n'.join(lines)

    def _submit_notification(self, message):
        libbe.util.subproc.invoke(self.notify, stdin=message, shell=True)


class ServeCommands (libbe.util.wsgi.ServerCommand):
    """Serve commands over HTTP.

    This allows you to run local `be` commands interfacing with remote
    data, transmitting command requests over the network.

    :py:class:`~libbe.command.base.Command` wrapper around
    :py:class:`ServerApp`.
    """

    name = 'serve-commands'

    def _get_app(self, logger, storage, **kwargs):
        return ServerApp(
            logger=logger, storage=storage, notify=kwargs.get('notify', False))

    def _long_help(self):
        return """
Example usage::

    $ be serve-commands

And in another terminal (or after backgrounding the server)::

    $ be --server http://localhost:8000/ list

If you bind your server to a public interface, take a look at the
``--read-only`` option or the combined ``--ssl --auth FILE``
options so other people can't mess with your repository.  If you do use
authentication, you'll need to send in your username and password::

    $ be --server http://username:password@localhost:8000/ list
"""


# alias for libbe.command.base.get_command_class()
Serve_commands = ServeCommands


if libbe.TESTING:
    class ServerAppTestCase (libbe.util.wsgi.WSGITestCase):
        def setUp(self):
            libbe.util.wsgi.WSGITestCase.setUp(self)
            self.bd = libbe.bugdir.SimpleBugDir(memory=False)
            self.app = ServerApp(self.bd.storage, logger=self.logger)

        def tearDown(self):
            self.bd.cleanup()
            libbe.util.wsgi.WSGITestCase.tearDown(self)

        def test_run_list(self):
            list = libbe.command.list.List()
            params = list._parse_options_args()
            data = libbe.storage.util.mapfile.generate({
                    'command': 'list',
                    'parameters': params,
                    }, context=0)
            self.getURL(self.app, '/run', method='POST', data=data)
            self.failUnless(self.status.startswith('200 '), self.status)
            self.failUnless(
                ('Content-Type', 'application/octet-stream'
                 ) in self.response_headers,
                self.response_headers)
            self.failUnless(self.exc_info == None, self.exc_info)
        # TODO: integration tests on ServeCommands?

    unitsuite =unittest.TestLoader().loadTestsFromModule(sys.modules[__name__])
    suite = unittest.TestSuite([unitsuite, doctest.DocTestSuite()])



# Copyright (C) 2009-2012 Chris Ball <cjb@laptop.org>
#                         Gianluca Montecchi <gian@grys.it>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

import sys

import libbe
import libbe.bugdir
import libbe.command
import libbe.command.util
import libbe.storage
import libbe.ui.util.editor


class Commit (libbe.command.Command):
    """Commit the currently pending changes to the repository

    >>> import sys
    >>> import libbe.bugdir
    >>> bd = libbe.bugdir.SimpleBugDir(memory=False, versioned=True)
    >>> io = libbe.command.StringInputOutput()
    >>> io.stdout = sys.stdout
    >>> ui = libbe.command.UserInterface(io=io)
    >>> ui.storage_callbacks.set_storage(bd.storage)
    >>> cmd = Commit(ui=ui)

    >>> bd.extra_strings = ['hi there']
    >>> bd.flush_reload()
    >>> ui.run(cmd, args=['Making a commit']) # doctest: +ELLIPSIS
    Committed ...
    >>> ui.cleanup()
    >>> bd.cleanup()
    """
    name = 'commit'

    def __init__(self, *args, **kwargs):
        libbe.command.Command.__init__(self, *args, **kwargs)
        self.options.extend([
                libbe.command.Option(name='body', short_name='b',
                    help='Provide the detailed body for the commit message.  In the special case that FILE == "EDITOR", spawn an editor to enter the body text (in which case you cannot use stdin for the summary)',
                    arg=libbe.command.Argument(name='body', metavar='FILE',
                        completion_callback=libbe.command.util.complete_path)),
                libbe.command.Option(name='allow-empty', short_name='a',
                                     help='Allow empty commits'),
                 ])
        self.args.extend([
                libbe.command.Argument(
                    name='summary', metavar='SUMMARY', default=None,
                    optional=True),
                ])

    def _run(self, **params):
        if params['summary'] == '-': # read summary from stdin
            assert params['body'] != 'EDITOR', \
                'Cannot spawn and editor when the summary is using stdin.'
            summary = sys.stdin.readline()
        else:
            summary = params['summary']
            if summary == None and params['body'] == None:
                params['body'] = 'EDITOR'
        storage = self._get_storage()
        if params['body'] == None:
            body = None
        elif params['body'] == 'EDITOR':
            body = libbe.ui.util.editor.editor_string(
                'Please enter your commit message above')
        else:
            self._check_restricted_access(storage, params['body'])
            body = libbe.util.encoding.get_file_contents(
                params['body'], decode=True)
        if summary == None:  # use the first body line as the summary
            if body == None:
                raise libbe.command.UserError(
                    'cannot commit without a summary')
            lines = body.splitlines()
            summary = lines[0]
            body = '\n'.join(lines[1:]).strip() + '\n'
        try:
            revision = storage.commit(summary, body=body,
                                      allow_empty=params['allow-empty'])
            print >> self.stdout, 'Committed %s' % revision
        except libbe.storage.EmptyCommit, e:
            print >> self.stdout, e
            return 1

    def _long_help(self):
        return """
Commit the current repository status.

The summary specified on the commandline is a string (only one line)
that describes the commit briefly or "-", in which case the string
will be read from stdin.  If no summary is given, the first line from
the body message is used instead.  If no summary or body is given, we
spawn an editor without needing the special "EDITOR" value for the
"--body" option.
"""



# Copyright (C) 2009-2012 Chris Ball <cjb@laptop.org>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

"""
Automatic pager for terminal output (a la Git).
"""

import sys, os, select

# see http://nex-3.com/posts/73-git-style-automatic-paging-in-ruby
def run_pager(paginate='auto'):
    """
    paginate should be one of 'never', 'auto', or 'always'.

    usage: just call this function and continue using sys.stdout like
    you normally would.
    """
    if paginate == 'never' \
            or sys.platform == 'win32' \
            or not hasattr(sys.stdout, 'isatty') \
            or sys.stdout.isatty() == False:
        return

    if paginate == 'auto':
        if 'LESS' not in os.environ:
            os.environ['LESS'] = '' # += doesn't work on undefined var
        # don't page if the input is short enough
        os.environ['LESS'] += ' -FRX'
    if 'PAGER' in os.environ:
        pager = os.environ['PAGER']
    else:
        pager = 'less'

    read_fd, write_fd = os.pipe()
    if os.fork() == 0:
        # child process
        os.close(read_fd)
        os.dup2(write_fd, 1)
        os.close(write_fd)
        if hasattr(sys.stderr, 'isatty') and sys.stderr.isatty() == True:
            os.dup2(1, 2)
        return

    # parent process, become pager
    os.close(write_fd)
    os.dup2(read_fd, 0)
    os.close(read_fd)

    # Wait until we have input before we start the pager
    select.select([0], [], [])
    os.execlp(pager, pager)



# Copyright (C) 2005-2012 Aaron Bentley <abentley@panoramicfeedback.com>
#                         Andrew Cooper <andrew.cooper@hkcreations.org>
#                         Chris Ball <cjb@laptop.org>
#                         Gianluca Montecchi <gian@grys.it>
#                         Niall Douglas (s_sourceforge@nedprod.com) <spam@spamtrap.com>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

import libbe
import libbe.command
import libbe.command.util

from .assign import parse_assigned as _parse_assigned


class New (libbe.command.Command):
    """Create a new bug

    >>> import os
    >>> import sys
    >>> import time
    >>> import libbe.bugdir
    >>> import libbe.util.id
    >>> bd = libbe.bugdir.SimpleBugDir(memory=False)
    >>> io = libbe.command.StringInputOutput()
    >>> io.stdout = sys.stdout
    >>> ui = libbe.command.UserInterface(io=io)
    >>> ui.storage_callbacks.set_storage(bd.storage)
    >>> cmd = New()

    >>> uuid_gen = libbe.util.id.uuid_gen
    >>> libbe.util.id.uuid_gen = lambda: 'X'
    >>> ui._user_id = u'Fran\\xe7ois'
    >>> options = {'assigned': 'none'}
    >>> ret = ui.run(cmd, options=options, args=['this is a test',])
    Created bug with ID abc/X
    >>> libbe.util.id.uuid_gen = uuid_gen
    >>> bd.flush_reload()
    >>> bug = bd.bug_from_uuid('X')
    >>> print bug.summary
    this is a test
    >>> bug.creator
    u'Fran\\xe7ois'
    >>> bug.reporter
    u'Fran\\xe7ois'
    >>> bug.time <= int(time.time())
    True
    >>> print bug.severity
    minor
    >>> print bug.status
    open
    >>> print bug.assigned
    None
    >>> ui.cleanup()
    >>> bd.cleanup()
    """
    name = 'new'

    def __init__(self, *args, **kwargs):
        libbe.command.Command.__init__(self, *args, **kwargs)
        self.options.extend([
                libbe.command.Option(name='reporter', short_name='r',
                    help='The user who reported the bug',
                    arg=libbe.command.Argument(
                        name='reporter', metavar='NAME')),
                libbe.command.Option(name='creator', short_name='c',
                    help='The user who created the bug',
                    arg=libbe.command.Argument(
                        name='creator', metavar='NAME')),
                libbe.command.Option(name='assigned', short_name='a',
                    help='The developer in charge of the bug',
                    arg=libbe.command.Argument(
                        name='assigned', metavar='NAME',
                        completion_callback=libbe.command.util.complete_assigned)),
                libbe.command.Option(name='status', short_name='t',
                    help='The bug\'s status level',
                    arg=libbe.command.Argument(
                        name='status', metavar='STATUS',
                        completion_callback=libbe.command.util.complete_status)),
                libbe.command.Option(name='severity', short_name='s',
                    help='The bug\'s severity',
                    arg=libbe.command.Argument(
                        name='severity', metavar='SEVERITY',
                        completion_callback=libbe.command.util.complete_severity)),
                libbe.command.Option(name='bugdir', short_name='b',
                    help='Short bugdir UUID for the new bug.  You '
                    'only need to set this if you have multiple bugdirs in '
                    'your repository.',
                    arg=libbe.command.Argument(
                        name='bugdir', metavar='ID', default=None,
                        completion_callback=libbe.command.util.complete_bugdir_id)),
                libbe.command.Option(name='full-uuid', short_name='f',
                    help='Print the full UUID for the new bug')
                ])
        self.args.extend([
                libbe.command.Argument(name='summary', metavar='SUMMARY')
                ])

    def _run(self, **params):
        if params['summary'] == '-': # read summary from stdin
            summary = self.stdin.readline()
        else:
            summary = params['summary']
        storage = self._get_storage()
        bugdirs = self._get_bugdirs()
        if params['bugdir']:
            bugdir = bugdirs[bugdir]
        elif len(bugdirs) == 1:
            bugdir = bugdirs.values()[0]
        else:
            raise libbe.command.UserError(
                'Ambiguous bugdir {}'.format(sorted(bugdirs.values())))
        storage.writeable = False
        bug = bugdir.new_bug(summary=summary.strip())
        if params['creator'] != None:
            bug.creator = params['creator']
        else:
            bug.creator = self._get_user_id()
        if params['reporter'] != None:
            bug.reporter = params['reporter']
        else:
            bug.reporter = bug.creator
        if params['assigned'] != None:
            bug.assigned = _parse_assigned(self, params['assigned'])
        if params['status'] != None:
            bug.status = params['status']
        if params['severity'] != None:
            bug.severity = params['severity']
        storage.writeable = True
        bug.save()
        if params['full-uuid']:
            bug_id = bug.id.long_user()
        else:
            bug_id = bug.id.user()
        self.stdout.write('Created bug with ID %s\n' % (bug_id))
        return 0

    def _long_help(self):
        return """
Create a new bug, with a new ID.  The summary specified on the
commandline is a string (only one line) that describes the bug briefly
or "-", in which case the string will be read from stdin.
"""



# Bugs Everywhere, a distributed bugtracker
# Copyright (C) 2008-2012 Chris Ball <cjb@laptop.org>
#                         Gianluca Montecchi <gian@grys.it>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

"""
Define editor_string(), a function that invokes an editor to accept
user-produced text as a string.
"""

import codecs
import locale
import os
import sys
import tempfile

import libbe
import libbe.util.encoding

if libbe.TESTING == True:
    import doctest


comment_marker = u"== Anything below this line will be ignored\n"

class CantFindEditor(Exception):
    def __init__(self):
        Exception.__init__(self, "Can't find editor to get string from")

def editor_string(comment=None, encoding=None):
    """Invokes the editor, and returns the user-produced text as a string

    >>> if "EDITOR" in os.environ:
    ...     del os.environ["EDITOR"]
    >>> if "VISUAL" in os.environ:
    ...     del os.environ["VISUAL"]
    >>> editor_string()
    Traceback (most recent call last):
    CantFindEditor: Can't find editor to get string from
    >>> os.environ["EDITOR"] = "echo bar > "
    >>> editor_string()
    u'bar\\n'
    >>> os.environ["VISUAL"] = "echo baz > "
    >>> editor_string()
    u'baz\\n'
    >>> os.environ["VISUAL"] = "echo 'baz\\n== Anything below this line will be ignored\\nHi' > "
    >>> editor_string()
    u'baz\\n'
    >>> del os.environ["EDITOR"]
    >>> del os.environ["VISUAL"]
    """
    if encoding == None:
        encoding = libbe.util.encoding.get_text_file_encoding()
    editor = None
    for name in ('VISUAL', 'EDITOR'):
        if name in os.environ and os.environ[name] != '':
            editor = os.environ[name]
            break
    if editor == None:
        raise CantFindEditor()
    fhandle, fname = tempfile.mkstemp()
    try:
        if comment is not None:
            cstring = u'\n'+comment_string(comment)
            os.write(fhandle, cstring.encode(encoding))
        os.close(fhandle)
        oldmtime = os.path.getmtime(fname)
        os.system("%s %s" % (editor, fname))
        output = libbe.util.encoding.get_file_contents(
            fname, encoding=encoding, decode=True)
        output = trimmed_string(output)
        if output.rstrip('\n') == "":
            output = None
    finally:
        os.unlink(fname)
    return output


def comment_string(comment):
    """
    >>> comment_string('hello') == comment_marker+"hello"
    True
    """
    return comment_marker + comment


def trimmed_string(instring):
    """
    >>> trimmed_string("hello\\n"+comment_marker)
    u'hello\\n'
    >>> trimmed_string("hi!\\n" + comment_string('Booga'))
    u'hi!\\n'
    """
    out = []
    for line in instring.splitlines(True):
        if line.startswith(comment_marker):
            break
        out.append(line)
    return ''.join(out)

if libbe.TESTING == True:
    suite = doctest.DocTestSuite()



# Copyright (C) 2005-2012 Aaron Bentley <abentley@panoramicfeedback.com>
#                         Chris Ball <cjb@laptop.org>
#                         Gianluca Montecchi <gian@grys.it>
#                         Oleg Romanyshyn <oromanyshyn@panoramicfeedback.com>
#                         Robert Lehmann <mail@robertlehmann.de>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

import itertools
import os
import re

import libbe
import libbe.bug
import libbe.command
import libbe.command.depend
from libbe.command.depend import Filter, parse_status, parse_severity
import libbe.command.tag
import libbe.command.target
import libbe.command.util

# get a list of * for cmp_*() comparing two bugs.
AVAILABLE_CMPS = [fn[4:] for fn in dir(libbe.bug) if fn[:4] == 'cmp_']
AVAILABLE_CMPS.remove('attr') # a cmp_* template.

class List (libbe.command.Command):
    """List bugs

    >>> import sys
    >>> import libbe.bugdir
    >>> bd = libbe.bugdir.SimpleBugDir(memory=False)
    >>> io = libbe.command.StringInputOutput()
    >>> io.stdout = sys.stdout
    >>> ui = libbe.command.UserInterface(io=io)
    >>> ui.storage_callbacks.set_storage(bd.storage)
    >>> cmd = List(ui=ui)

    >>> ret = ui.run(cmd)
    abc/a:om: Bug A
    >>> ret = ui.run(cmd, {'status':'closed'})
    abc/b:cm: Bug B
    >>> ret = ui.run(cmd, {'status':'all', 'sort':'time'})
    abc/a:om: Bug A
    abc/b:cm: Bug B
    >>> bd.storage.writeable
    True
    >>> ui.cleanup()
    >>> bd.cleanup()
    """

    name = 'list'

    def __init__(self, *args, **kwargs):
        libbe.command.Command.__init__(self, *args, **kwargs)
        self.options.extend([
                libbe.command.Option(name='status',
                    help='Only show bugs matching the STATUS specifier',
                    arg=libbe.command.Argument(
                        name='status', metavar='STATUS', default='active',
                        completion_callback=libbe.command.util.complete_status)),
                libbe.command.Option(name='severity',
                    help='Only show bugs matching the SEVERITY specifier',
                    arg=libbe.command.Argument(
                        name='severity', metavar='SEVERITY', default='all',
                        completion_callback=libbe.command.util.complete_severity)),
                libbe.command.Option(name='important',
                    help='List bugs with >= "serious" severity'),
                libbe.command.Option(name='assigned', short_name='a',
                    help='Only show bugs matching ASSIGNED',
                    arg=libbe.command.Argument(
                        name='assigned', metavar='ASSIGNED', default=None,
                        completion_callback=libbe.command.util.complete_assigned)),
                libbe.command.Option(name='mine', short_name='m',
                    help='List bugs assigned to you'),
                libbe.command.Option(name='extra-strings', short_name='e',
                    help='Only show bugs matching STRINGS, e.g. --extra-strings'
                         ' TAG:working,TAG:xml',
                    arg=libbe.command.Argument(
                        name='extra-strings', metavar='STRINGS', default=None,
                        completion_callback=libbe.command.util.complete_extra_strings)),
                libbe.command.Option(name='sort', short_name='S',
                    help='Adjust bug-sort criteria with comma-separated list '
                         'SORT.  e.g. "--sort creator,time".  '
                         'Available criteria: %s' % ','.join(AVAILABLE_CMPS),
                    arg=libbe.command.Argument(
                        name='sort', metavar='SORT', default=None,
                        completion_callback=libbe.command.util.Completer(AVAILABLE_CMPS))),
                libbe.command.Option(name='tags', short_name='t',
                    help='Add TAGS: field to standard listing format.'),
                libbe.command.Option(name='ids', short_name='i',
                    help='Only print the bug IDS'),
                libbe.command.Option(name='xml', short_name='x',
                    help='Dump output in XML format'),
                ])
#    parser.add_option("-S", "--sort", metavar="SORT-BY", dest="sort_by",
#                      help="Adjust bug-sort criteria with comma-separated list SORT-BY.  e.g. \"--sort creator,time\".  Available criteria: %s" % ','.join(AVAILABLE_CMPS), default=None)
#    # boolean options.  All but ids and xml are special cases of long forms
#             ("w", "wishlist", "List bugs with 'wishlist' severity"),
#             ("A", "active", "List all active bugs"),
#             ("U", "unconfirmed", "List unconfirmed bugs"),
#             ("o", "open", "List open bugs"),
#             ("T", "test", "List bugs in testing"),
#    for s in bools:
#        attr = s[1].replace('-','_')
#        short = "-%c" % s[0]
#        long = "--%s" % s[1]
#        help = s[2]
#        parser.add_option(short, long, action="store_true",
#                          dest=attr, help=help, default=False)
#    return parser
#
#                ])

    def _run(self, **params):
        storage = self._get_storage()
        bugdirs = self._get_bugdirs()
        writeable = storage.writeable
        storage.writeable = False
        cmp_list, status, severity, assigned, extra_strings_regexps = \
            self._parse_params(bugdirs, params)
        filter = Filter(status, severity, assigned,
                        extra_strings_regexps=extra_strings_regexps)
        bugs = list(itertools.chain(*list(
                    [bugdir.bug_from_uuid(uuid) for uuid in bugdir.uuids()]
                    for bugdir in bugdirs.values())))
        bugs = [b for b in bugs if filter(bugdirs, b) == True]
        self.result = bugs
        if len(bugs) == 0 and params['xml'] == False:
            print >> self.stdout, 'No matching bugs found'

        # sort bugs
        bugs = self._sort_bugs(bugs, cmp_list)

        # print list of bugs
        if params['ids'] == True:
            for bug in bugs:
                print >> self.stdout, bug.id.user()
        else:
            self._list_bugs(bugs, show_tags=params['tags'], xml=params['xml'])
        storage.writeable = writeable
        return 0

    def _parse_params(self, bugdirs, params):
        cmp_list = []
        if params['sort'] != None:
            for cmp in params['sort'].split(','):
                if cmp not in AVAILABLE_CMPS:
                    raise libbe.command.UserError(
                        'Invalid sort on "%s".\nValid sorts:\n  %s'
                    % (cmp, '\n  '.join(AVAILABLE_CMPS)))
                cmp_list.append(getattr(libbe.bug, 'cmp_%s' % cmp))
        status = parse_status(params['status'])
        severity = parse_severity(params['severity'],
                                  important=params['important'])
        # select assigned
        if params['assigned'] == None:
            if params['mine'] == True:
                assigned = [self._get_user_id()]
            else:
                assigned = 'all'
        else:
            assigned = libbe.command.util.select_values(
                params['assigned'], libbe.command.util.assignees(bugdirs))
        for i in range(len(assigned)):
            if assigned[i] == '-':
                assigned[i] = params['user-id']
        if params['extra-strings'] == None:
            extra_strings_regexps = []
        else:
            extra_strings_regexps = [re.compile(x)
                                     for x in params['extra-strings'].split(',')]
        return (cmp_list, status, severity, assigned, extra_strings_regexps)

    def _sort_bugs(self, bugs, cmp_list=None):
        if cmp_list is None:
            cmp_list = []
        cmp_list.extend(libbe.bug.DEFAULT_CMP_FULL_CMP_LIST)
        cmp_fn = libbe.bug.BugCompoundComparator(cmp_list=cmp_list)
        bugs.sort(cmp_fn)
        return bugs

    def _list_bugs(self, bugs, show_tags=False, xml=False):
        if xml == True:
            print >> self.stdout, \
                '<?xml version="1.0" encoding="%s" ?>' % self.stdout.encoding
            print >> self.stdout, '<be-xml>'
        if len(bugs) > 0:
            for bug in bugs:
                if xml == True:
                    print >> self.stdout, bug.xml(show_comments=True)
                else:
                    bug_string = bug.string(shortlist=True)
                    if show_tags == True:
                        attrs,summary = bug_string.split(' ', 1)
                        bug_string = (
                            '%s%s: %s'
                            % (attrs,
                               ','.join(libbe.command.tag.get_tags(bug)),
                               summary))
                    print >> self.stdout, bug_string
        if xml == True:
            print >> self.stdout, '</be-xml>'

    def _long_help(self):
        return """
This command lists bugs.  Normally it prints a short string like
  bea/576:om:[TAGS:] Allow attachments
Where
  bea/576   the bug id
  o         the bug status is 'open' (first letter)
  m         the bug severity is 'minor' (first letter)
  TAGS      comma-separated list of bug tags (if --tags is set)
  Allo...   the bug summary string

You can optionally (-u) print only the bug ids.

There are several criteria that you can filter by:
  * status
  * severity
  * assigned (who the bug is assigned to)
Allowed values for each criterion may be given in a comma seperated
list.  The special string "all" may be used with any of these options
to match all values of the criterion.  As with the --status and
--severity options for `be depend`, starting the list with a minus
sign makes your selections a blacklist instead of the default
whitelist.

status
  %s
severity
  %s
assigned
  free form, with the string '-' being a shortcut for yourself.

In addition, there are some shortcut options that set boolean flags.
The boolean options are ignored if the matching string option is used.
""" % (','.join(libbe.bug.status_values),
       ','.join(libbe.bug.severity_values))



# Copyright (C) 2009-2012 Chris Ball <cjb@laptop.org>
#                         Valtteri Kokkoniemi <rvk@iki.fi>
#                         W. Trevor King <wking@tremily.us>
#
# This file is part of Bugs Everywhere.
#
# Bugs Everywhere is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation, either version 2 of the License, or (at your option) any
# later version.
#
# Bugs Everywhere is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# Bugs Everywhere.  If not, see <http://www.gnu.org/licenses/>.

import copy
import os
import sys
try: # import core module, Python >= 2.5
    from xml.etree import ElementTree
except ImportError: # look for non-core module
    from elementtree import ElementTree

import libbe
import libbe.bug
import libbe.bugdir
import libbe.command
import libbe.command.util
import libbe.comment
import libbe.util.encoding
import libbe.util.id
import libbe.util.utility

if libbe.TESTING == True:
    import doctest
    import StringIO
    import unittest

    import libbe.bugdir


class Import_XML (libbe.command.Command):
    """Import comments and bugs from XML

    >>> import time
    >>> import StringIO
    >>> import libbe.bugdir
    >>> bd = libbe.bugdir.SimpleBugDir(memory=False)
    >>> io = libbe.command.StringInputOutput()
    >>> io.stdout = sys.stdout
    >>> ui = libbe.command.UserInterface(io=io)
    >>> ui.storage_callbacks.set_storage(bd.storage)
    >>> cmd = Import_XML(ui=ui)

    >>> ui.io.set_stdin('<be-xml><comment><uuid>c</uuid><body>This is a comment about a</body></comment></be-xml>')
    >>> ret = ui.run(cmd, {'root':'/a'}, ['-'])
    >>> bd.flush_reload()
    >>> bug = bd.bug_from_uuid('a')
    >>> bug.load_comments(load_full=False)
    >>> comment = bug.comment_root[0]
    >>> print comment.body
    This is a comment about a
    <BLANKLINE>
    >>> comment.time <= int(time.time())
    True
    >>> comment.in_reply_to is None
    True
    >>> ui.cleanup()
    >>> bd.cleanup()
    """
    name = 'import-xml'

    def __init__(self, *args, **kwargs):
        libbe.command.Command.__init__(self, *args, **kwargs)
        self.options.extend([
                libbe.command.Option(name='ignore-missing-references', short_name='i',
                    help="If any comment's <in-reply-to> refers to a non-existent comment, ignore it (instead of raising an exception)."),
                libbe.command.Option(name='add-only', short_name='a',
                    help='If any bug or comment listed in the XML file already exists in the bug repository, do not alter the repository version.'),
                libbe.command.Option(name='preserve-uuids', short_name='p',
                    help='Preserve UUIDs for trusted input (potential name collisions).'),
                libbe.command.Option(name='root', short_name='r',
                    help='Supply a bugdir, bug, or comment ID as the root of '
                    'any non-bugdir elements that are direct children of the '
                    '<be-xml> element.  If any such elements exist, you are '
                    'required to set this option.',
                    arg=libbe.command.Argument(
                        name='root', metavar='ID',
                        completion_callback=libbe.command.util.complete_bug_comment_id)),
                ])
        self.args.extend([
                libbe.command.Argument(
                    name='xml-file', metavar='XML-FILE'),
                ])

    def _run(self, **params):
        storage = self._get_storage()
        bugdirs = self._get_bugdirs()
        writeable = storage.writeable
        storage.writeable = False
        if params['root'] != None:
            root_bugdir,root_bug,root_comment = (
                libbe.command.util.bugdir_bug_comment_from_user_id(
                    bugdirs, params['root']))
        else:
            root_bugdir,root_bug,root_comment = (None, None, None)

        xml = self._read_xml(storage, params)
        version,root_bugdirs,root_bugs,root_comments = self._parse_xml(
            xml, params)

        if params['add-only']:
            accept_changes = False
            accept_extra_strings = False
        else:
            accept_changes = True
            accept_extra_strings = True

        dirty_items = list(self._merge_comments(
                bugdirs, root_bug, root_comment, root_comments,
                params, accept_changes, accept_extra_strings))
        dirty_items.extend(self._merge_bugs(
                bugdirs, root_bugdir, root_bugs,
                params, accept_changes, accept_extra_strings))
        dirty_items.extend(self._merge_bugdirs(
                bugdirs, root_bugdirs,
                params, accept_changes, accept_extra_strings))

        # protect against programmer error causing data loss:
        if root_bug is not None:
            # check for each of the new comments
            comms = []
            for c in root_bug.comments():
                comms.append(c.uuid)
                if c.alt_id != None:
                    comms.append(c.alt_id)
            if root_comment.uuid == libbe.comment.INVALID_UUID:
                root_text = root_bug.id.user()
            else:
                root_text = root_comment.id.user()
            for new in root_comments:
                assert new.uuid in comms or new.alt_id in comms, \
                    "comment %s (alt: %s) wasn't added to %s" \
                    % (new.uuid, new.alt_id, root_text)
        for new in root_bugs:
            # check for each of the new bugs
            try:
                libbe.command.util.bug_from_uuid(bugdirs, new.uuid)
            except libbe.bugdir.NoBugMatches:
                try:
                    libbe.command.util.bug_from_uuid(bugdirs, new.alt_id)
                except libbe.bugdir.NoBugMatches:
                    raise AssertionError(
                        "bug {0} (alt: {1}) wasn't added to {2}".format(
                            new.uuid, new.alt_id, root_bugdir.id.user()))
        for new in root_bugdirs:
            assert new.uuid in bugdirs or new.alt_id in bugdirs, (
                "bugdir {0} wasn't added to {1}".format(
                    new.uuid, sorted(bugdirs.keys())))

        # save new information
        storage.writeable = writeable
        for item in dirty_items:
            item.save()

    def _read_xml(self, storage, params):
        if params['xml-file'] == '-':
            return self.stdin.read().encode(self.stdin.encoding)
        else:
            self._check_restricted_access(storage, params['xml-file'])
            return libbe.util.encoding.get_file_contents(params['xml-file'])

    def _parse_xml(self, xml, params):
        version = {}
        root_bugdirs = []
        root_bugs = []
        root_comments = []
        be_xml = ElementTree.XML(xml)
        if be_xml.tag != 'be-xml':
            raise libbe.util.utility.InvalidXML(
                'import-xml', be_xml, 'root element must be <be-xml>')
        for child in be_xml.getchildren():
            if child.tag == 'bugdir':
                new = libbe.bugdir.BugDir(storage=None)
                new.from_xml(child, preserve_uuids=params['preserve-uuids'])
                root_bugdirs.append(new)
            elif child.tag == 'bug':
                new = libbe.bug.Bug()
                new.from_xml(child, preserve_uuids=params['preserve-uuids'])
                root_bugs.append(new)
            elif child.tag == 'comment':
                new = libbe.comment.Comment()
                new.from_xml(child, preserve_uuids=params['preserve-uuids'])
                root_comments.append(new)
            elif child.tag == 'version':
                for gchild in child.getchildren():
                    if child.tag in ['tag', 'nick', 'revision', 'revision-id']:
                        text = xml.sax.saxutils.unescape(child.text)
                        text = text.decode('unicode_escape').strip()
                        version[child.tag] = text
                    else:
                        sys.stderr.write(
                            'ignoring unknown tag {0} in {1}\n'.format(
                                gchild.tag, child.tag))
            else:
                sys.stderr.write('ignoring unknown tag {0} in {1}\n'.format(
                        child.tag, be_xml.tag))
        return (version, root_bugdirs, root_bugs, root_comments)

    def _merge_comments(self, bugdirs, bug, root_comment, comments,
                        params, accept_changes, accept_extra_strings,
                        accept_comments=True):
        if len(comments) == 0:
            return
        if bug is None:
            raise libbe.command.UserError(
                'No root bug for merging comments:\n{0}'.format(
                    '\n\n'.join([c.string() for c in comments])))
        bug.load_comments(load_full=True)
        if root_comment.uuid == libbe.comment.INVALID_UUID:
            root_comment = bug.comment_root
        else:
            root_comment = bug.comment_from_uuid(root_comment.uuid)
        new_bug = libbe.bug.Bug(bugdir=bug.bugdir, uuid=bug.uuid)
        new_bug.explicit_attrs = []
        new_bug.comment_root = copy.deepcopy(bug.comment_root)
        if root_comment.uuid == libbe.comment.INVALID_UUID:
            new_root_comment = new_bug.comment_root
        else:
            new_root_comment = new_bug.comment_from_uuid(
                root_comment.uuid)
        for new in new_bug.comments():
            new.explicit_attrs = []
        try:
            new_bug.add_comments(
                comments,
                default_parent=root_comment,
                ignore_missing_references=params['ignore-missing-references'])
        except libbe.comment.MissingReference as e:
            raise libbe.command.UserError(e)
        bug.merge(new_bug, accept_changes=accept_changes,
                  accept_extra_strings=accept_extra_strings,
                  accept_comments=accept_comments)
        yield bug

    def _merge_bugs(self, bugdirs, bugdir, bugs,
                    params, accept_changes, accept_extra_strings,
                    accept_comments=True):
        for new in bugs:
            try:
                old = bugdir.bug_from_uuid(new.alt_id)
            except KeyError:
                bugdir.append(new, update=True)
                yield new
            else:
                old.load_comments(load_full=True)
                old.merge(new, accept_changes=accept_changes,
                          accept_extra_strings=accept_extra_strings,
                          accept_comments=accept_comments)
                yield old

    def _merge_bugdirs(self, bugdirs, new_bugdirs,
                       params, accept_changes, accept_extra_strings,
                       accept_comments=True):
        for new in new_bugdirs:
            if new.alt_id in bugdirs:
                old = bugdirs[new.alt_id]
                old.load_all_bugs()
                old.merge(new, accept_changes=accept_changes,
                          accept_extra_strings=accept_extra_strings,
                          accept_bugs=True,
                          accept_comments=accept_comments)
                yield old
            else:
                bugdirs[new.uuid] = new
                new.storage = self._get_storage()
                yield new

    def _long_help(self):
        return """
Import comments and bugs from XMLFILE.  If XMLFILE is '-', the file is
read from stdin.

This command provides a fallback mechanism for passing bugs between
repositories, in case the repositories VCSs are incompatible.  If the
VCSs are compatible, it's better to use their builtin merge/push/pull
to share this information, as that will preserve a more detailed
history.

The XML file should be formatted similarly to:

  <be-xml>
    <version>
      <tag>1.0.0</tag>
      <branch-nick>be</branch-nick>
      <revno>446</revno>
      <revision-id>a@b.com-20091119214553-iqyw2cpqluww3zna</revision-id>
    <version>
    <bugdir>
      <bug>
        ...
        <comment>...</comment>
        <comment>...</comment>
      </bug>
      <bug>...</bug>
    </bugdir>
    <bug>...</bug>
    <bug>...</bug>
    <comment>...</comment>
    <comment>...</comment>
  </be-xml>

where the ellipses mark output commpatible with BugDir.xml(),
Bug.xml(), and Comment.xml().  Take a look at the output of `be show
--xml` for some explicit examples.  Unrecognized tags are ignored.
Missing tags are left at the default value.  The version tag is not
required, but is strongly recommended.

The bugdir, bug, and comment UUIDs are always auto-generated, so if
you set a <uuid> field, but no <alt-id> field, your <uuid> will be
used as the object's <alt-id>.  An exception is raised if <alt-id>
conflicts with an existing object.  Bugdirs and bugs do not have a
permantent alt-id, so they the <uuid>s you specify are not saved.  The
<uuid>s _are_ used to match agains prexisting bug and comment uuids,
and comment alt-ids, and fields explicitly given in the XML file will
replace old versions unless the --add-only flag.

*.extra_strings recieves special treatment, and if --add-only is not
set, the resulting list concatenates both source lists and removes
repeats.

Here's an example of import activity:
  Repository
   bugdir (uuid=abc123)
     bug (uuid=B, creator=John, status=open)
       estr (don't forget your towel)
       estr (helps with space travel)
       com (uuid=C1, author=Jane, body=Hello)
       com (uuid=C2, author=Jess, body=World)
  XML
   bugdir (uuid=abc123)
     bug (uuid=B, status=fixed)
       estr (don't forget your towel)
       estr (watch out for flying dolphins)
       com (uuid=C1, body=So long)
       com (uuid=C3, author=Jed, body=And thanks)
  Result
   bugdir (uuid=abc123)
     bug (uuid=B, creator=John, status=fixed)
       estr (don't forget your towel)
       estr (helps with space travel)
       estr (watch out for flying dolphins)
       com (uuid=C1, author=Jane, body=So long)
       com (uuid=C2, author=Jess, body=World)
       com (uuid=C4, alt-id=C3, author=Jed, body=And thanks)
  Result, with --add-only
   bugdir (uuid=abc123)
     bug (uuid=B, creator=John, status=open)
       estr (don't forget your towel)
       estr (helps with space travel)
       com (uuid=C1, author=Jane, body=Hello)
       com (uuid=C2, author=Jess, body=World)
       com (uuid=C4, alt-id=C3, author=Jed, body=And thanks)

Examples:

Import comments (e.g. emails from a mailbox) and append to bug /XYZ:

  $ be-mail-to-xml mail.mbox | be import-xml -r /XYZ -

Or you can append those emails underneath the prexisting comment /XYZ/3:

  $ be-mail-to-xml mail.mbox | be import-xml -r /XYZ/3 -

User creates a new bug:

  user$ be new "The demuxulizer is broken"
  Created bug with ID 48f
  user$ be comment 48f
  <Describe bug>
  ...

User exports bug as xml and emails it to the developers:

  user$ be show --xml 48f > 48f.xml
  user$ cat 48f.xml | mail -s "Demuxulizer bug xml" devs@b.com
or equivalently (with a slightly fancier be-handle-mail compatible
email):
  user$ be email-bugs 48f

Devs recieve email, and save it's contents as demux-bug.xml:

  dev$ cat demux-bug.xml | be import-xml -
"""


Import_xml = Import_XML # alias for libbe.command.base.get_command_class()

if libbe.TESTING == True:
    class LonghelpTestCase (unittest.TestCase):
        """
        Test import scenarios given in longhelp.
        """
        def setUp(self):
            self.bugdir = libbe.bugdir.SimpleBugDir(memory=False)
            io = libbe.command.StringInputOutput()
            self.ui = libbe.command.UserInterface(io=io)
            self.ui.storage_callbacks.set_storage(self.bugdir.storage)
            self.cmd = Import_XML(ui=self.ui)
            self.cmd._storage = self.bugdir.storage
            self.cmd._setup_io = lambda i_enc,o_enc : None
            bugA = self.bugdir.bug_from_uuid('a')
            self.bugdir.remove_bug(bugA)
            self.bugdir.storage.writeable = False
            bugB = self.bugdir.bug_from_uuid('b')
            bugB.creator = 'John'
            bugB.status = 'open'
            bugB.extra_strings += ["don't forget your towel"]
            bugB.extra_strings += ['helps with space travel']
            comm1 = bugB.comment_root.new_reply(body='Hello\n')
            comm1.uuid = 'c1'
            comm1.author = 'Jane'
            comm2 = bugB.comment_root.new_reply(body='World\n')
            comm2.uuid = 'c2'
            comm2.author = 'Jess'
            self.bugdir.storage.writeable = True
            bugB.save()
            self.xml = """
            <be-xml>
              <bugdir>
                <uuid>abc123</uuid>
                <bug>
                  <uuid>b</uuid>
                  <status>fixed</status>
                  <summary>a test bug</summary>
                  <extra-string>don't forget your towel</extra-string>
                  <extra-string>watch out for flying dolphins</extra-string>
                  <comment>
                    <uuid>c1</uuid>
                    <body>So long</body>
                  </comment>
                  <comment>
                    <uuid>c3</uuid>
                    <author>Jed</author>
                    <body>And thanks</body>
                  </comment>
                </bug>
              </bugdir>
            </be-xml>
            """
            self.root_comment_xml = """
            <be-xml>
              <comment>
                <uuid>c1</uuid>
                <body>So long</body>
              </comment>
              <comment>
                <uuid>c3</uuid>
                <author>Jed</author>
                <body>And thanks</body>
              </comment>
            </be-xml>
            """
        def tearDown(self):
            self.bugdir.cleanup()
            self.ui.cleanup()
        def _execute(self, xml, params={}, args=[]):
            self.ui.io.set_stdin(xml)
            self.ui.run(self.cmd, params, args)
            self.bugdir.flush_reload()
        def testCleanBugdir(self):
            uuids = list(self.bugdir.uuids())
            self.failUnless(uuids == ['b'], uuids)
        def testNotAddOnly(self):
            bugB = self.bugdir.bug_from_uuid('b')
            self._execute(self.xml, {}, ['-'])
            uuids = list(self.bugdir.uuids())
            self.failUnless(uuids == ['b'], uuids)
            bugB = self.bugdir.bug_from_uuid('b')
            self.failUnless(bugB.uuid == 'b', bugB.uuid)
            self.failUnless(bugB.creator == 'John', bugB.creator)
            self.failUnless(bugB.status == 'fixed', bugB.status)
            self.failUnless(bugB.summary == 'a test bug', bugB.summary)
            estrs = ["don't forget your towel",
                     'helps with space travel',
                     'watch out for flying dolphins']
            self.failUnless(bugB.extra_strings == estrs, bugB.extra_strings)
            comments = list(bugB.comments())
            self.failUnless(len(comments) == 3,
                            ['%s (%s, %s)' % (c.uuid, c.alt_id, c.body)
                             for c in comments])
            c1 = bugB.comment_from_uuid('c1')
            comments.remove(c1)
            self.failUnless(c1.uuid == 'c1', c1.uuid)
            self.failUnless(c1.alt_id == None, c1.alt_id)
            self.failUnless(c1.author == 'Jane', c1.author)
            self.failUnless(c1.body == 'So long\n', c1.body)
            c2 = bugB.comment_from_uuid('c2')
            comments.remove(c2)
            self.failUnless(c2.uuid == 'c2', c2.uuid)
            self.failUnless(c2.alt_id == None, c2.alt_id)
            self.failUnless(c2.author == 'Jess', c2.author)
            self.failUnless(c2.body == 'World\n', c2.body)
            c4 = comments[0]
            self.failUnless(len(c4.uuid) == 36, c4.uuid)
            self.failUnless(c4.alt_id == 'c3', c4.alt_id)
            self.failUnless(c4.author == 'Jed', c4.author)
            self.failUnless(c4.body == 'And thanks\n', c4.body)
        def testAddOnly(self):
            bugB = self.bugdir.bug_from_uuid('b')
            initial_bugB_summary = bugB.summary
            self._execute(self.xml, {'add-only':True}, ['-'])
            uuids = list(self.bugdir.uuids())
            self.failUnless(uuids == ['b'], uuids)
            bugB = self.bugdir.bug_from_uuid('b')
            self.failUnless(bugB.uuid == 'b', bugB.uuid)
            self.failUnless(bugB.creator == 'John', bugB.creator)
            self.failUnless(bugB.status == 'open', bugB.status)
            self.failUnless(bugB.summary == initial_bugB_summary, bugB.summary)
            estrs = ["don't forget your towel",
                     'helps with space travel']
            self.failUnless(bugB.extra_strings == estrs, bugB.extra_strings)
            comments = list(bugB.comments())
            self.failUnless(len(comments) == 3,
                            ['%s (%s)' % (c.uuid, c.alt_id) for c in comments])
            c1 = bugB.comment_from_uuid('c1')
            comments.remove(c1)
            self.failUnless(c1.uuid == 'c1', c1.uuid)
            self.failUnless(c1.alt_id == None, c1.alt_id)
            self.failUnless(c1.author == 'Jane', c1.author)
            self.failUnless(c1.body == 'Hello\n', c1.body)
            c2 = bugB.comment_from_uuid('c2')
            comments.remove(c2)
            self.failUnless(c2.uuid == 'c2', c2.uuid)
            self.failUnless(c2.alt_id == None, c2.alt_id)
            self.failUnless(c2.author == 'Jess', c2.author)
            self.failUnless(c2.body == 'World\n', c2.body)
            c4 = comments[0]
            self.failUnless(len(c4.uuid) == 36, c4.uuid)
            self.failUnless(c4.alt_id == 'c3', c4.alt_id)
            self.failUnless(c4.author == 'Jed', c4.author)
            self.failUnless(c4.body == 'And thanks\n', c4.body)
        def testRootCommentsNotAddOnly(self):
            bugB = self.bugdir.bug_from_uuid('b')
            initial_bugB_summary = bugB.summary
            self._execute(self.root_comment_xml, {'root':'/b'}, ['-'])
            uuids = list(self.bugdir.uuids())
            uuids = list(self.bugdir.uuids())
            self.failUnless(uuids == ['b'], uuids)
            bugB = self.bugdir.bug_from_uuid('b')
            self.failUnless(bugB.uuid == 'b', bugB.uuid)
            self.failUnless(bugB.creator == 'John', bugB.creator)
            self.failUnless(bugB.status == 'open', bugB.status)
            self.failUnless(bugB.summary == initial_bugB_summary, bugB.summary)
            estrs = ["don't forget your towel",
                     'helps with space travel']
            self.failUnless(bugB.extra_strings == estrs, bugB.extra_strings)
            comments = list(bugB.comments())
            self.failUnless(len(comments) == 3,
                            ['%s (%s, %s)' % (c.uuid, c.alt_id, c.body)
                             for c in comments])
            c1 = bugB.comment_from_uuid('c1')
            comments.remove(c1)
            self.failUnless(c1.uuid == 'c1', c1.uuid)
            self.failUnless(c1.alt_id == None, c1.alt_id)
            self.failUnless(c1.author == 'Jane', c1.author)
            self.failUnless(c1.body == 'So long\n', c1.body)
            c2 = bugB.comment_from_uuid('c2')
            comments.remove(c2)
            self.failUnless(c2.uuid == 'c2', c2.uuid)
            self.failUnless(c2.alt_id == None, c2.alt_id)
            self.failUnless(c2.author == 'Jess', c2.author)
            self.failUnless(c2.body == 'World\n', c2.body)
            c4 = comments[0]
            self.failUnless(len(c4.uuid) == 36, c4.uuid)
            self.failUnless(c4.alt_id == 'c3', c4.alt_id)
            self.failUnless(c4.author == 'Jed', c4.author)
            self.failUnless(c4.body == 'And thanks\n', c4.body)
        def testRootCommentsAddOnly(self):
            bugB = self.bugdir.bug_from_uuid('b')
            initial_bugB_summary = bugB.summary
            self._execute(self.root_comment_xml,
                          {'root':'/b', 'add-only':True}, ['-'])
            uuids = list(self.bugdir.uuids())
            self.failUnless(uuids == ['b'], uuids)
            bugB = self.bugdir.bug_from_uuid('b')
            self.failUnless(bugB.uuid == 'b', bugB.uuid)
            self.failUnless(bugB.creator == 'John', bugB.creator)
            self.failUnless(bugB.status == 'open', bugB.status)
            self.failUnless(bugB.summary == initial_bugB_summary, bugB.summary)
            estrs = ["don't forget your towel",
                     'helps with space travel']
            self.failUnless(bugB.extra_strings == estrs, bugB.extra_strings)
            comments = list(bugB.comments())
            self.failUnless(len(comments) == 3,
                            ['%s (%s)' % (c.uuid, c.alt_id) for c in comments])
            c1 = bugB.comment_from_uuid('c1')
            comments.remove(c1)
            self.failUnless(c1.uuid == 'c1', c1.uuid)
            self.failUnless(c1.alt_id == None, c1.alt_id)
            self.failUnless(c1.author == 'Jane', c1.author)
            self.failUnless(c1.body == 'Hello\n', c1.body)
            c2 = bugB.comment_from_uuid('c2')
            comments.remove(c2)
            self.failUnless(c2.uuid == 'c2', c2.uuid)
            self.failUnless(c2.alt_id == None, c2.alt_id)
            self.failUnless(c2.author == 'Jess', c2.author)
            self.failUnless(c2.body == 'World\n', c2.body)
            c4 = comments[0]
            self.failUnless(len(c4.uuid) == 36, c4.uuid)
            self.failUnless(c4.alt_id == 'c3', c4.alt_id)
            self.failUnless(c4.author == 'Jed', c4.author)
            self.failUnless(c4.body == 'And thanks\n', c4.body)

    unitsuite =unittest.TestLoader().loadTestsFromModule(sys.modules[__name__])
    suite = unittest.TestSuite([unitsuite, doctest.DocTestSuite()])
